@@ -1,8 +1,6 @@
 /* 
    neon SSL/TLS support using OpenSSL
-   Copyright (C) 2002-2009, Joe Orton <joe@manyfish.co.uk>
-   Portions are:
-   Copyright (C) 1999-2000 Tommi Komulainen <Tommi.Komulainen@iki.fi>
+   Copyright (C) 2002-2011, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -36,17 +34,21 @@
 #include <openssl/pkcs12.h>
 #include <openssl/x509v3.h>
 #include <openssl/rand.h>
+#include <openssl/opensslv.h>
+#include <openssl/evp.h>
 
 #ifdef NE_HAVE_TS_SSL
 #include <stdlib.h> /* for abort() */
+#ifndef _WIN32
 #include <pthread.h>
 #endif
+#endif
 
 #include "ne_ssl.h"
 #include "ne_string.h"
 #include "ne_session.h"
 #include "ne_internal.h"
-
+#include "ne_md5.h"
 #include "ne_private.h"
 #include "ne_privssl.h"
 
@@ -65,6 +67,15 @@ typedef unsigned char ne_d2i_uchar;
 typedef const unsigned char ne_d2i_uchar;
 #endif
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define X509_up_ref(x) x->references++
+#define EVP_PKEY_up_ref(x) x->references++
+#define EVP_MD_CTX_new() ne_calloc(sizeof(EVP_MD_CTX))
+#define EVP_MD_CTX_free(ctx) ne_free(ctx)
+#define EVP_MD_CTX_reset EVP_MD_CTX_cleanup
+#define EVP_PKEY_get0_RSA(evp) (evp->pkey.rsa)
+#endif
+
 struct ne_ssl_dname_s {
     X509_NAME *dn;
 };
@@ -84,6 +95,8 @@ struct ne_ssl_client_cert_s {
     char *friendly_name;
 };
 
+#define NE_SSL_UNHANDLED (0x20) /* failure bit for unhandled case. */
+
 /* Append an ASN.1 DirectoryString STR to buffer BUF as UTF-8.
  * Returns zero on success or non-zero on error. */
 static int append_dirstring(ne_buffer *buf, ASN1_STRING *str)
@@ -95,7 +108,7 @@ static int append_dirstring(ne_buffer *buf, ASN1_STRING *str)
     case V_ASN1_IA5STRING: /* definitely ASCII */
     case V_ASN1_VISIBLESTRING: /* probably ASCII */
     case V_ASN1_PRINTABLESTRING: /* subset of ASCII */
-        ne__buffer_qappend(buf, str->data, str->length);
+        ne_buffer_qappend(buf, str->data, str->length);
         break;
     case V_ASN1_UTF8STRING:
         /* Fail for embedded NUL bytes. */
@@ -136,7 +149,7 @@ static int append_dirstring(ne_buffer *buf, ASN1_STRING *str)
  * safety. */
 static char *dup_ia5string(const ASN1_IA5STRING *as)
 {
-    return ne__strnqdup(as->data, as->length);
+    return ne_strnqdup(as->data, as->length);
 }
 
 char *ne_ssl_readable_dname(const ne_ssl_dname *name)
@@ -148,15 +161,16 @@ char *ne_ssl_readable_dname(const ne_ssl_dname *name)
 
     for (n = X509_NAME_entry_count(name->dn); n > 0; n--) {
 	X509_NAME_ENTRY *ent = X509_NAME_get_entry(name->dn, n-1);
+	ASN1_OBJECT *obj = X509_NAME_ENTRY_get_object(ent);
 	
         /* Skip commonName or emailAddress except if there is no other
          * attribute in dname. */
-	if ((OBJ_cmp(ent->object, cname) && OBJ_cmp(ent->object, email)) ||
+	if ((OBJ_cmp(obj, cname) && OBJ_cmp(obj, email)) ||
             (!flag && n == 1)) {
  	    if (flag++)
 		ne_buffer_append(dump, ", ", 2);
 
-            if (append_dirstring(dump, ent->value))
+            if (append_dirstring(dump, X509_NAME_ENTRY_get_data(ent)))
                 ne_buffer_czappend(dump, "???");
 	}
     }
@@ -353,6 +367,62 @@ static ne_ssl_certificate *populate_cert(ne_ssl_certificate *cert, X509 *x5)
     return cert;
 }
 
+/* OpenSSL cert verification callback.  This is invoked for *each*
+ * error which is encoutered whilst verifying the cert chain; multiple
+ * invocations for any particular cert in the chain are possible. */
+static int verify_callback(int ok, X509_STORE_CTX *ctx)
+{
+    /* OpenSSL, living in its own little happy world of global state,
+     * where userdata was just a twinkle in the eye of an API designer
+     * yet to be born.  Or... "Seriously, wtf?"  */
+    SSL *ssl = X509_STORE_CTX_get_ex_data(ctx, 
+                                          SSL_get_ex_data_X509_STORE_CTX_idx());
+    ne_session *sess = SSL_get_app_data(ssl);
+    int depth = X509_STORE_CTX_get_error_depth(ctx);
+    int err = X509_STORE_CTX_get_error(ctx);
+    int failures = 0;
+
+    /* If there's no error, nothing to do here. */
+    if (ok) return ok;
+
+    NE_DEBUG(NE_DBG_SSL, "ssl: Verify callback @ %d => %d\n", depth, err);
+
+    /* Map the error code onto any of the exported cert validation
+     * errors, if possible. */
+    switch (err) {
+    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
+    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
+    case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
+    case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
+    case X509_V_ERR_CERT_UNTRUSTED:
+    case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
+        failures |= NE_SSL_UNTRUSTED;
+        break;
+    case X509_V_ERR_CERT_NOT_YET_VALID:
+        failures |= depth > 0 ? NE_SSL_BADCHAIN : NE_SSL_NOTYETVALID;
+        break;
+    case X509_V_ERR_CERT_HAS_EXPIRED:
+        failures |= depth > 0 ? NE_SSL_BADCHAIN : NE_SSL_EXPIRED;
+        break;
+    case X509_V_OK:
+        break;
+    default:
+        /* Clear the failures bitmask so check_certificate knows this
+         * is a bailout. */
+        sess->ssl_context->failures |= NE_SSL_UNHANDLED;
+        NE_DEBUG(NE_DBG_SSL, "ssl: Unhandled verification error %d -> %s\n", 
+                 err, X509_verify_cert_error_string(err));
+        return 0;
+    }
+
+    sess->ssl_context->failures |= failures;
+
+    NE_DEBUG(NE_DBG_SSL, "ssl: Verify failures |= %d => %d\n", failures,
+             sess->ssl_context->failures);
+    
+    return 1;
+}
+
 /* Return a linked list of certificate objects from an OpenSSL chain. */
 static ne_ssl_certificate *make_chain(STACK_OF(X509) *chain)
 {
@@ -385,17 +455,21 @@ static ne_ssl_certificate *make_chain(STACK_OF(X509) *chain)
 static int check_certificate(ne_session *sess, SSL *ssl, ne_ssl_certificate *chain)
 {
     X509 *cert = chain->subject;
-    ASN1_TIME *notBefore = X509_get_notBefore(cert);
-    ASN1_TIME *notAfter = X509_get_notAfter(cert);
-    int ret, failures = 0;
-    long result;
+    int ret, failures = sess->ssl_context->failures;
     ne_uri server;
 
-    /* check expiry dates */
-    if (X509_cmp_current_time(notBefore) >= 0)
-	failures |= NE_SSL_NOTYETVALID;
-    else if (X509_cmp_current_time(notAfter) <= 0)
-	failures |= NE_SSL_EXPIRED;
+    /* If the verification callback hit a case which can't be mapped
+     * to one of the exported error bits, it's treated as a hard
+     * failure rather than invoking the callback, which can't present
+     * a useful error to the user.  "Um, something is wrong.  OK?" */
+    if (failures & NE_SSL_UNHANDLED) {
+        long result = SSL_get_verify_result(ssl);
+
+        ne_set_error(sess, _("Certificate verification error: %s"),
+                    X509_verify_cert_error_string(result));
+
+        return NE_ERROR;
+    }
 
     /* Check certificate was issued to this server; pass URI of
      * server. */
@@ -409,38 +483,6 @@ static int check_certificate(ne_session *sess, SSL *ssl, ne_ssl_certificate *cha
         return NE_ERROR;
     } else if (ret > 0) failures |= NE_SSL_IDMISMATCH;
 
-    /* get the result of the cert verification out of OpenSSL */
-    result = SSL_get_verify_result(ssl);
-
-    NE_DEBUG(NE_DBG_SSL, "Verify result: %ld = %s\n", result,
-	     X509_verify_cert_error_string(result));
-
-    switch (result) {
-    case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
-    case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
-    case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
-	/* TODO: and probably more result codes here... */
-	failures |= NE_SSL_UNTRUSTED;
-	break;
-	/* ignore these, since we've already noticed them: */
-    case X509_V_ERR_CERT_NOT_YET_VALID:
-    case X509_V_ERR_CERT_HAS_EXPIRED:
-        /* cert was trusted: */
-    case X509_V_OK:
-	break;
-    default:
-	/* TODO: tricky to handle the 30-odd failure cases OpenSSL
-	 * presents here (see x509_vfy.h), and present a useful API to
-	 * the application so it in turn can then present a meaningful
-	 * UI to the user.  The only thing to do really would be to
-	 * pass back the error string, but that's not localisable.  So
-	 * just fail the verification here - better safe than
-	 * sorry. */
-	ne_set_error(sess, _("Certificate verification error: %s"),
-		     X509_verify_cert_error_string(result));
-	return NE_ERROR;
-    }
-
     if (failures == 0) {
         /* verified OK! */
         ret = NE_OK;
@@ -469,8 +511,8 @@ static ne_ssl_client_cert *dup_client_cert(const ne_ssl_client_cert *cc)
 
     populate_cert(&newcc->cert, cc->cert.subject);
 
-    cc->cert.subject->references++;
-    cc->pkey->references++;
+    X509_up_ref(cc->cert.subject);
+    EVP_PKEY_up_ref(cc->pkey);
     return newcc;
 }
 
@@ -480,7 +522,7 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
     ne_session *const sess = SSL_get_app_data(ssl);
 
     if (!sess->client_cert && sess->ssl_provide_fn) {
-	ne_ssl_dname **dnames = NULL;
+	ne_ssl_dname **dnames = NULL, *dnarray = NULL;
         int n, count = 0;
 	STACK_OF(X509_NAME) *ca_list = SSL_get_client_CA_list(ssl);
 
@@ -488,9 +530,10 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
 
         if (count > 0) {
             dnames = ne_malloc(count * sizeof(ne_ssl_dname *));
+            dnarray = ne_malloc(count * sizeof(ne_ssl_dname));
             
             for (n = 0; n < count; n++) {
-                dnames[n] = ne_malloc(sizeof(ne_ssl_dname));
+                dnames[n] = &dnarray[n];
                 dnames[n]->dn = sk_X509_NAME_value(ca_list, n);
             }
         }
@@ -499,8 +542,7 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
 	sess->ssl_provide_fn(sess->ssl_provide_ud, sess, 
                              (const ne_ssl_dname *const *)dnames, count);
         if (count) {
-            for (n = 0; n < count; n++)
-                ne_free(dnames[n]);
+            ne_free(dnarray);
             ne_free(dnames);
         }
     }
@@ -508,8 +550,8 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
     if (sess->client_cert) {
         ne_ssl_client_cert *const cc = sess->client_cert;
 	NE_DEBUG(NE_DBG_SSL, "Supplying client certificate.\n");
-	cc->pkey->references++;
-	cc->cert.subject->references++;
+	EVP_PKEY_up_ref(cc->pkey);
+	X509_up_ref(cc->cert.subject);
 	*cert = cc->cert.subject;
 	*pkey = cc->pkey;
 	return 1;
@@ -535,12 +577,21 @@ ne_ssl_context *ne_ssl_context_create(int mode)
         SSL_CTX_set_client_cert_cb(ctx->ctx, provide_client_cert);
         /* enable workarounds for buggy SSL server implementations */
         SSL_CTX_set_options(ctx->ctx, SSL_OP_ALL);
+        SSL_CTX_set_verify(ctx->ctx, SSL_VERIFY_PEER, verify_callback);
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+        SSL_CTX_set_post_handshake_auth(ctx->ctx, 1);
+#endif
     } else if (mode == NE_SSL_CTX_SERVER) {
         ctx->ctx = SSL_CTX_new(SSLv23_server_method());
         SSL_CTX_set_session_cache_mode(ctx->ctx, SSL_SESS_CACHE_CLIENT);
+#ifdef SSL_OP_NO_TICKET
+        /* disable ticket support since it inhibits testing of session
+         * caching. */
+        SSL_CTX_set_options(ctx->ctx, SSL_OP_NO_TICKET);
+#endif
     } else {
-        ctx->ctx = SSL_CTX_new(SSLv2_server_method());
-        SSL_CTX_set_session_cache_mode(ctx->ctx, SSL_SESS_CACHE_CLIENT);
+        ne_free(ctx);
+        return NULL;
     }
     return ctx;
 }
@@ -564,6 +615,22 @@ void ne_ssl_context_set_flag(ne_ssl_context *ctx, int flag, int value)
     SSL_CTX_set_options(ctx->ctx, opts);
 }
 
+int ne_ssl_context_get_flag(ne_ssl_context *ctx, int flag)
+{
+    switch (flag) {
+    case NE_SSL_CTX_SSLv2:
+#ifdef OPENSSL_NO_SSL2
+        return 0;
+#else
+        return ! (SSL_CTX_get_options(ctx->ctx) & SSL_OP_NO_SSLv2);
+#endif
+    default:
+        break;
+    }
+
+    return 0;
+}
+
 int ne_ssl_context_keypair(ne_ssl_context *ctx, const char *cert,
                            const char *key)
 {
@@ -571,7 +638,7 @@ int ne_ssl_context_keypair(ne_ssl_context *ctx, const char *cert,
 
     ret = SSL_CTX_use_PrivateKey_file(ctx->ctx, key, SSL_FILETYPE_PEM);
     if (ret == 1) {
-        ret = SSL_CTX_use_certificate_file(ctx->ctx, cert, SSL_FILETYPE_PEM);
+        ret = SSL_CTX_use_certificate_chain_file(ctx->ctx, cert);
     }
 
     return ret == 1 ? 0 : -1;
@@ -604,6 +671,25 @@ void ne_ssl_context_destroy(ne_ssl_context *ctx)
     ne_free(ctx);
 }
 
+#if !defined(HAVE_SSL_SESSION_CMP) && !defined(SSL_SESSION_cmp) \
+    && defined(OPENSSL_VERSION_NUMBER) \
+    && OPENSSL_VERSION_NUMBER > 0x10000000L
+/* OpenSSL 1.0 removed SSL_SESSION_cmp for no apparent reason - hoping
+ * it is reasonable to assume that comparing the session IDs is
+ * sufficient. */
+static int SSL_SESSION_cmp(SSL_SESSION *a, SSL_SESSION *b)
+{
+    const unsigned char *session1_buf, *session2_buf;
+    unsigned int session1_len, session2_len;
+
+    session1_buf = SSL_SESSION_get_id(a, &session1_len);
+    session2_buf = SSL_SESSION_get_id(b, &session2_len);
+
+    return session1_len == session2_len
+        && memcmp(session1_buf, session2_buf, session1_len) == 0;
+}
+#endif
+
 /* For internal use only. */
 int ne__negotiate_ssl(ne_session *sess)
 {
@@ -619,6 +705,7 @@ int ne__negotiate_ssl(ne_session *sess)
         sess->flags[NE_SESSFLAG_TLS_SNI] ? sess->server.hostname : NULL;
 
     sess->ssl_cc_requested = 0;
+    ctx->failures = 0;
 
     if (ne_sock_connect_ssl(sess->socket, ctx, sess)) {
 	if (ctx->sess) {
@@ -627,12 +714,12 @@ int ne__negotiate_ssl(ne_session *sess)
 	    ctx->sess = NULL;
 	}
         if (sess->ssl_cc_requested) {
-            ne_set_error(sess, _("SSL negotiation failed, "
+            ne_set_error(sess, _("SSL handshake failed, "
                                  "client certificate was requested: %s"),
                          ne_sock_error(sess->socket));
         }
         else {
-            ne_set_error(sess, _("SSL negotiation failed: %s"),
+            ne_set_error(sess, _("SSL handshake failed: %s"),
                          ne_sock_error(sess->socket));
         }
         return NE_ERROR;
@@ -656,17 +743,10 @@ int ne__negotiate_ssl(ne_session *sess)
 	return NE_ERROR;
     }
 
-    if (sess->server_cert) {
-        int diff = X509_cmp(sk_X509_value(chain, 0), sess->server_cert->subject);
+    if (sess->server_cert 
+        && X509_cmp(sk_X509_value(chain, 0), sess->server_cert->subject) == 0) {
+        /* Same leaf cert used as last time - no need to reverify. */
         if (freechain) sk_X509_free(chain); /* no longer need the chain */
-	if (diff) {
-	    /* This could be a MITM attack: fail the request. */
-	    ne_set_error(sess, _("Server certificate changed: "
-				 "connection intercepted?"));
-	    return NE_ERROR;
-	} 
-	/* certificate has already passed verification: no need to
-	 * verify it again. */
     } else {
 	/* new connection: create the chain. */
         ne_ssl_certificate *cert = make_chain(chain);
@@ -770,22 +850,12 @@ static char *find_friendly_name(PKCS12 *p12)
     return name;
 }
 
-ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
+static ne_ssl_client_cert *parse_client_cert(PKCS12 *p12)
 {
-    PKCS12 *p12;
-    FILE *fp;
     X509 *cert;
     EVP_PKEY *pkey;
     ne_ssl_client_cert *cc;
 
-    fp = fopen(filename, "rb");
-    if (fp == NULL)
-        return NULL;
-
-    p12 = d2i_PKCS12_fp(fp, NULL);
-
-    fclose(fp);
-    
     if (p12 == NULL) {
         ERR_clear_error();
         return NULL;
@@ -831,6 +901,34 @@ ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
     }
 }
 
+ne_ssl_client_cert *ne_ssl_clicert_import(const unsigned char *buffer, 
+                                          size_t buflen)
+{
+    ne_d2i_uchar *p;
+    PKCS12 *p12;
+
+    p = buffer;
+    p12 = d2i_PKCS12(NULL, &p, buflen);
+    
+    return parse_client_cert(p12);
+}
+    
+ne_ssl_client_cert *ne_ssl_clicert_read(const char *filename)
+{
+    PKCS12 *p12;
+    FILE *fp;
+
+    fp = fopen(filename, "rb");
+    if (fp == NULL)
+        return NULL;
+
+    p12 = d2i_PKCS12_fp(fp, NULL);
+
+    fclose(fp);
+
+    return parse_client_cert(p12);
+}
+
 #ifdef HAVE_PAKCHOIS
 ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
                                                  size_t der_len,
@@ -839,8 +937,8 @@ ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
     ne_ssl_client_cert *cc;
     ne_d2i_uchar *p;
     X509 *x5;
-    RSA *pk;    
-    EVP_PKEY *epk, *tpk;
+    EVP_PKEY *pubkey, *privkey;
+    RSA *rsa;
 
     p = der;
     x5 = d2i_X509(NULL, &p, der_len); /* p is incremented */
@@ -848,24 +946,28 @@ ne_ssl_client_cert *ne__ssl_clicert_exkey_import(const unsigned char *der,
         ERR_clear_error();
         return NULL;
     }
-    
-    pk = RSA_new();
-    RSA_set_method(pk, method);
-    epk = EVP_PKEY_new();
-    EVP_PKEY_assign_RSA(epk, pk);
-    
-    /* It is necessary to initialize pk->n otherwise OpenSSL will barf
-     * later calling RSA_size() on this RSA structure.
-     * X509_get_pubkey() forces the relevant RSA parameters to be
-     * extracted from the certificate. */
-    tpk = X509_get_pubkey(x5);
-    pk->n = BN_dup(tpk->pkey.rsa->n);
-    EVP_PKEY_free(tpk);
 
-    cc = ne_calloc(sizeof *cc);
+    pubkey = X509_get_pubkey(x5);
+    if (EVP_PKEY_base_id(pubkey) != EVP_PKEY_RSA) {
+        X509_free(x5);
+        NE_DEBUG(NE_DBG_SSL, "ssl: Only RSA private keys are supported via PKCS#11.\n");
+        return NULL;
+    }
+
+    /* Duplicate the public parameters of the RSA key. */
+    rsa = RSAPublicKey_dup(EVP_PKEY_get0_RSA(pubkey));
+    /* Done with the copied public key. */
+    EVP_PKEY_free(pubkey);
     
+    /* Switch to using customer RSA_METHOD for RSA object. */
+    RSA_set_method(rsa, method);
+    /* Set up new EVP_PKEY. */
+    privkey = EVP_PKEY_new();
+    EVP_PKEY_assign_RSA(privkey, rsa);
+    
+    cc = ne_calloc(sizeof *cc);
     cc->decrypted = 1;
-    cc->pkey = epk;
+    cc->pkey = privkey;
 
     populate_cert(&cc->cert, x5);
 
@@ -1035,7 +1137,10 @@ int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
     return 0;
 }
 
-#ifdef NE_HAVE_TS_SSL
+#if defined(NE_HAVE_TS_SSL) && OPENSSL_VERSION_NUMBER < 0x10100000L
+/* From OpenSSL 1.1.0 locking callbacks are no longer neede. */
+#define WITH_OPENSSL_LOCKING (1)
+
 /* Implementation of locking callbacks to make OpenSSL thread-safe.
  * If the OpenSSL API was better designed, this wouldn't be necessary.
  * In OpenSSL releases without CRYPTO_set_idptr_callback, it's not
@@ -1043,17 +1148,25 @@ int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
  * it's necessary to cast from a pthread_t to an unsigned long at some
  * point.  */
 
+#ifndef _WIN32
 static pthread_mutex_t *locks;
+#else
+static HANDLE *locks;
+#endif
 static size_t num_locks;
 
 #ifndef HAVE_CRYPTO_SET_IDPTR_CALLBACK
 /* Named to be obvious when it shows up in a backtrace. */
 static unsigned long thread_id_neon(void)
 {
+#ifndef _WIN32
     /* This will break if pthread_t is a structure; upgrading OpenSSL
      * >= 0.9.9 (which does not require this callback) is the only
      * solution.  */
     return (unsigned long) pthread_self();
+#else
+    return (unsigned long) GetCurrentThreadId();
+#endif
 }
 #endif
 
@@ -1062,19 +1175,25 @@ static unsigned long thread_id_neon(void)
 static void thread_lock_neon(int mode, int n, const char *file, int line)
 {
     if (mode & CRYPTO_LOCK) {
+#ifndef _WIN32
         if (pthread_mutex_lock(&locks[n])) {
+#else
+        if (WaitForSingleObject(locks[n], INFINITE)) {
+#endif
             abort();
         }
     }
     else {
+#ifndef _WIN32
         if (pthread_mutex_unlock(&locks[n])) {
+#else
+        if (!ReleaseMutex(locks[n])) {
+#endif
             abort();
         }
     }
 }
 
-#endif
-
 /* ID_CALLBACK_IS_{NEON,OTHER} evaluate as true if the currently
  * registered OpenSSL ID callback is the neon function (_NEON), or has
  * been overwritten by some other app (_OTHER). */
@@ -1085,15 +1204,18 @@ static void thread_lock_neon(int mode, int n, const char *file, int line)
 #define ID_CALLBACK_IS_OTHER (CRYPTO_get_id_callback() != NULL)
 #define ID_CALLBACK_IS_NEON (CRYPTO_get_id_callback() == thread_id_neon)
 #endif
+        
+#endif /* NE_HAVE_TS_SSL && OPENSSL_VERSION_NUMBER < 1.1.1 */
 
 int ne__ssl_init(void)
 {
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     CRYPTO_malloc_init();
     SSL_load_error_strings();
     SSL_library_init();
     OpenSSL_add_all_algorithms();
 
-#ifdef NE_HAVE_TS_SSL
+#ifdef WITH_OPENSSL_LOCKING
     /* If some other library has already come along and set up the
      * thread-safety callbacks, then it must be presumed that the
      * other library will have a longer lifetime in the process than
@@ -1116,7 +1238,11 @@ int ne__ssl_init(void)
 
         locks = malloc(num_locks * sizeof *locks);
         for (n = 0; n < num_locks; n++) {
+#ifndef _WIN32
             if (pthread_mutex_init(&locks[n], NULL)) {
+#else
+            if ((locks[n] = CreateMutex(NULL, FALSE, NULL)) == NULL) {
+#endif
                 NE_DEBUG(NE_DBG_SOCKET, "ssl: Failed to initialize pthread mutex.\n");
                 return -1;
             }
@@ -1126,6 +1252,7 @@ int ne__ssl_init(void)
                  "for %" NE_FMT_SIZE_T " locks.\n", num_locks);
     }
 #endif
+#endif /* OPENSSL_VERSION_NUMBER < 0x10100000L */
 
     return 0;
 }
@@ -1135,7 +1262,7 @@ void ne__ssl_exit(void)
     /* Cannot call ERR_free_strings() etc here in case any other code
      * in the process using OpenSSL. */
 
-#ifdef NE_HAVE_TS_SSL
+#ifdef WITH_OPENSSL_LOCKING
     /* Only unregister the callbacks if some *other* library has not
      * come along in the mean-time and trampled over the callbacks
      * installed by neon. */
@@ -1149,10 +1276,83 @@ void ne__ssl_exit(void)
         CRYPTO_set_locking_callback(NULL);
 
         for (n = 0; n < num_locks; n++) {
+#ifndef _WIN32
             pthread_mutex_destroy(&locks[n]);
+#else
+            CloseHandle(locks[n]);
+#endif
         }
 
         free(locks);
     }
 #endif
 }
+
+struct ne_md5_ctx {
+    EVP_MD_CTX *ctx;
+};
+
+/* Returns zero on succes, non-zero on failure. */
+static int init_md5_ctx(struct ne_md5_ctx *ctx)
+{
+    ctx->ctx = EVP_MD_CTX_new();
+
+    if (EVP_DigestInit_ex(ctx->ctx, EVP_md5(), NULL) != 1) {
+        return 1;
+    }
+
+    return 0;
+}
+
+struct ne_md5_ctx *ne_md5_create_ctx(void)
+{
+    struct ne_md5_ctx *ctx = ne_malloc(sizeof *ctx);
+    
+    if (init_md5_ctx(ctx)) {
+        ne_free(ctx);
+        return NULL;
+    }
+    
+    return ctx;
+}
+
+void ne_md5_process_block(const void *buffer, size_t len,
+                          struct ne_md5_ctx *ctx)
+{
+    EVP_DigestUpdate(ctx->ctx, buffer, len);
+}
+
+void ne_md5_process_bytes(const void *buffer, size_t len,
+                          struct ne_md5_ctx *ctx)
+{
+    EVP_DigestUpdate(ctx->ctx, buffer, len);
+}
+
+void *ne_md5_finish_ctx(struct ne_md5_ctx *ctx, void *resbuf)
+{
+    EVP_DigestFinal(ctx->ctx, resbuf, NULL);
+    
+    return resbuf;
+}
+
+struct ne_md5_ctx *ne_md5_dup_ctx(struct ne_md5_ctx *ctx)
+{
+    struct ne_md5_ctx *r = ne_md5_create_ctx();
+
+    EVP_MD_CTX_copy_ex(r->ctx, ctx->ctx);
+    
+    return r;
+}
+
+void ne_md5_reset_ctx(struct ne_md5_ctx *ctx)
+{
+    EVP_MD_CTX_reset(ctx->ctx);
+
+    init_md5_ctx(ctx);    
+}
+    
+void ne_md5_destroy_ctx(struct ne_md5_ctx *ctx)
+{
+    EVP_MD_CTX_free(ctx->ctx);
+    ne_free(ctx);
+}
