@@ -1,6 +1,6 @@
 /* 
    WebDAV property manipulation
-   Copyright (C) 2000-2003, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2000-2008, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -33,6 +33,10 @@
 #include "ne_props.h"
 #include "ne_basic.h"
 #include "ne_locks.h"
+#include "ne_internal.h"
+
+/* don't store flat props with a value > 10K */
+#define MAX_FLATPROP_LEN (102400)
 
 struct ne_propfind_handler_s {
     ne_session *sess;
@@ -45,23 +49,22 @@ struct ne_propfind_handler_s {
     ne_207_parser *parser207;
     ne_xml_parser *parser;
 
-    /* Callback to create the private structure. */
-    ne_props_create_complex private_creator;
-    void *private_userdata;
+    /* Creator/destructor callbacks. */
+    ne_props_create_complex creator;
+    ne_props_destroy_complex destructor;
+    void *cd_userdata;
     
-    /* Current propset. */
+    /* Current propset, or NULL if none being processed. */
     ne_prop_result_set *current;
 
+    ne_buffer *value; /* current flat property value */
+    int depth; /* nesting depth within a flat property */
+
     ne_props_result callback;
     void *userdata;
 };
 
-#define ELM_namedprop (NE_ELM_207_UNUSED)
-
-static const struct ne_xml_elm flat_elms[] = {
-    { "", "", NE_ELM_unknown, NE_XML_COLLECT },
-    { NULL }
-};
+#define ELM_flatprop (NE_207_STATE_TOP - 1)
 
 /* We build up the results of one 'response' element in memory. */
 struct prop {
@@ -82,18 +85,29 @@ struct propstat {
 /* Results set. */
 struct ne_prop_result_set_s {
     struct propstat *pstats;
-    int numpstats;
+    int numpstats, counter;
     void *private;
-    char *href;
+    ne_uri uri;
 };
 
+#define MAX_PROP_COUNTER (1024)
 
 static int 
-startelm(void *userdata, const struct ne_xml_elm *elm, 
+startelm(void *userdata, int state, const char *name, const char *nspace,
 	 const char **atts);
 static int 
-endelm(void *userdata, const struct ne_xml_elm *elm, const char *cdata);
-static int check_context(void *ud, ne_xml_elmid parent, ne_xml_elmid child);
+endelm(void *userdata, int state, const char *name, const char *nspace);
+
+/* Handle character data; flat property value. */
+static int chardata(void *userdata, int state, const char *data, size_t len)
+{
+    ne_propfind_handler *hdl = userdata;
+
+    if (state == ELM_flatprop && hdl->value->length < MAX_FLATPROP_LEN)
+        ne_buffer_append(hdl->value, data, len);
+
+    return 0;
+}
 
 ne_xml_parser *ne_propfind_get_parser(ne_propfind_handler *handler)
 {
@@ -113,13 +127,8 @@ static int propfind(ne_propfind_handler *handler,
 
     /* Register the flat property handler to catch any properties 
      * which the user isn't handling as 'complex'. */
-    ne_xml_push_handler(handler->parser, flat_elms, 
-			 check_context, startelm, endelm, handler);
+    ne_xml_push_handler(handler->parser, startelm, chardata, endelm, handler);
 
-    /* Register the catch-all handler to ignore any cruft the
-     * server returns. */
-    ne_207_ignore_unknown(handler->parser207);
-    
     handler->callback = results;
     handler->userdata = userdata;
 
@@ -135,8 +144,8 @@ static int propfind(ne_propfind_handler *handler,
 
     if (ret == NE_OK && ne_get_status(req)->klass != 2) {
 	ret = NE_ERROR;
-    } else if (!ne_xml_valid(handler->parser)) {
-	ne_set_error(handler->sess, ne_xml_get_error(handler->parser));
+    } else if (ne_xml_failed(handler->parser)) {
+	ne_set_error(handler->sess, "%s", ne_xml_get_error(handler->parser));
 	ret = NE_ERROR;
     }
 
@@ -149,13 +158,13 @@ static void set_body(ne_propfind_handler *hdl, const ne_propname *names)
     int n;
     
     if (!hdl->has_props) {
-	ne_buffer_zappend(body, "<prop>" EOL);
+	ne_buffer_czappend(body, "<prop>\n");
 	hdl->has_props = 1;
     }
 
     for (n = 0; names[n].name != NULL; n++) {
 	ne_buffer_concat(body, "<", names[n].name, " xmlns=\"", 
-			 NSPACE(names[n].nspace), "\"/>" EOL, NULL);
+			 NSPACE(names[n].nspace), "\"/>\n", NULL);
     }
 
 }
@@ -163,7 +172,7 @@ static void set_body(ne_propfind_handler *hdl, const ne_propname *names)
 int ne_propfind_allprop(ne_propfind_handler *handler, 
 			 ne_props_result results, void *userdata)
 {
-    ne_buffer_zappend(handler->body, "<allprop/></propfind>" EOL);
+    ne_buffer_czappend(handler->body, "<allprop/></propfind>\n");
     return propfind(handler, results, userdata);
 }
 
@@ -171,7 +180,7 @@ int ne_propfind_named(ne_propfind_handler *handler, const ne_propname *props,
 		       ne_props_result results, void *userdata)
 {
     set_body(handler, props);
-    ne_buffer_zappend(handler->body, "</prop></propfind>" EOL);
+    ne_buffer_czappend(handler->body, "</prop></propfind>\n");
     return propfind(handler, results, userdata);
 }
 
@@ -185,8 +194,8 @@ int ne_proppatch(ne_session *sess, const char *uri,
     int n, ret;
     
     /* Create the request body */
-    ne_buffer_zappend(body, "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" EOL
-		     "<D:propertyupdate xmlns:D=\"DAV:\">");
+    ne_buffer_czappend(body, "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"
+                       "<D:propertyupdate xmlns:D=\"DAV:\">");
 
     for (n = 0; items[n].name != NULL; n++) {
 	const char *elm = (items[n].type == ne_propset) ? "set" : "remove";
@@ -205,16 +214,16 @@ int ne_proppatch(ne_session *sess, const char *uri,
 	    ne_buffer_append(body, ">", 1);
 	}
 
-	ne_buffer_concat(body, "</", items[n].name->name, "></D:prop></D:", elm, ">"
-			 EOL, NULL);
+	ne_buffer_concat(body, "</", items[n].name->name, "></D:prop></D:", elm, 
+                         ">\n", NULL);
     }	
 
-    ne_buffer_zappend(body, "</D:propertyupdate>" EOL);
+    ne_buffer_czappend(body, "</D:propertyupdate>\n");
 
     ne_set_request_body_buffer(req, body->data, ne_buffer_size(body));
     ne_add_request_header(req, "Content-Type", NE_XML_MEDIA_TYPE);
     
-#ifdef USE_DAV_LOCKS
+#ifdef NE_HAVE_DAV
     ne_lock_using_resource(req, uri, NE_DEPTH_ZERO);
 #endif
 
@@ -292,7 +301,7 @@ const char *ne_propset_lang(const ne_prop_result_set *set,
 
 void *ne_propfind_current_private(ne_propfind_handler *handler)
 {
-    return handler->current->private;
+    return handler->current ? handler->current->private : NULL;
 }
 
 void *ne_propset_private(const ne_prop_result_set *set)
@@ -335,25 +344,15 @@ const ne_status *ne_propset_status(const ne_prop_result_set *set,
     }
 }
 
-
-/* Pick up all properties as flat properties. */
-static int check_context(void *ud, ne_xml_elmid parent, ne_xml_elmid child)
-{
-    if (child == NE_ELM_unknown && parent == NE_ELM_prop)
-	return NE_XML_VALID;
-
-    return NE_XML_DECLINE;
-}
-
-static void *start_response(void *userdata, const char *href)
+static void *start_response(void *userdata, const ne_uri *uri)
 {
     ne_prop_result_set *set = ne_calloc(sizeof(*set));
     ne_propfind_handler *hdl = userdata;
 
-    set->href = ne_strdup(href);
+    ne_uri_copy(&set->uri, uri);
 
-    if (hdl->private_creator != NULL) {
-	set->private = hdl->private_creator(hdl->private_userdata, href);
+    if (hdl->creator) {
+	set->private = hdl->creator(hdl->cd_userdata, &set->uri);
     }
 
     hdl->current = set;
@@ -364,9 +363,15 @@ static void *start_response(void *userdata, const char *href)
 static void *start_propstat(void *userdata, void *response)
 {
     ne_prop_result_set *set = response;
-    int n;
+    ne_propfind_handler *hdl = userdata;
     struct propstat *pstat;
+    int n;
 
+    if (++hdl->current->counter == MAX_PROP_COUNTER) {
+        ne_xml_set_error(hdl->parser, _("Response exceeds maximum property count"));
+        return NULL;
+    }
+    
     n = set->numpstats;
     set->pstats = ne_realloc(set->pstats, sizeof(struct propstat) * (n+1));
     set->numpstats = n+1;
@@ -378,9 +383,8 @@ static void *start_propstat(void *userdata, void *response)
     return &set->pstats[n];
 }
 
-static int 
-startelm(void *userdata, const struct ne_xml_elm *elm, 
-	 const char **atts)
+static int startelm(void *userdata, int parent,
+                    const char *nspace, const char *name, const char **atts)
 {
     ne_propfind_handler *hdl = userdata;
     struct propstat *pstat = ne_207_get_current_propstat(hdl->parser207);
@@ -388,10 +392,48 @@ startelm(void *userdata, const struct ne_xml_elm *elm,
     int n;
     const char *lang;
 
-    /* Paranoia */
-    if (pstat == NULL) {
-	NE_DEBUG(NE_DBG_XML, "gp_startelm: No propstat found, or not my element.");
-	return -1;
+    /* Just handle all children of propstat and their descendants. */
+    if ((parent != NE_207_STATE_PROP && parent != ELM_flatprop) 
+        || pstat == NULL)
+        return NE_XML_DECLINE;
+
+    if (parent == ELM_flatprop) {
+        /* collecting the flatprop value. */
+        hdl->depth++;
+        if (hdl->value->used < MAX_FLATPROP_LEN) {
+            const char **a = atts;
+
+            ne_buffer_concat(hdl->value, "<", nspace, name, NULL);
+            
+            while (a[0] && hdl->value->used < MAX_FLATPROP_LEN) {
+                const char *nsep = strchr(a[0], ':'), *pfx;
+
+                /* Resolve the attribute namespace prefix, if any.
+                 * Ignore a failure to resolve the namespace prefix. */
+                pfx = nsep ? ne_xml_resolve_nspace(hdl->parser,
+                                                   a[0], nsep - a[0]) : NULL;
+                
+                if (pfx) {
+                    ne_buffer_concat(hdl->value, " ", pfx, nsep + 1, "='", 
+                                     a[1], "'", NULL);
+                }
+                else {
+                    ne_buffer_concat(hdl->value, " ", a[0], "='", a[1], "'", NULL);
+                }
+                a += 2;
+            }
+
+            ne_buffer_czappend(hdl->value, ">");
+        }
+
+        return ELM_flatprop;
+    }        
+
+    /* Enforce maximum number of properties per resource to prevent a
+     * memory exhaustion attack by a hostile server. */
+    if (++hdl->current->counter == MAX_PROP_COUNTER) {
+        ne_xml_set_error(hdl->parser, _("Response exceeds maximum property count"));
+        return NE_XML_ABORT;
     }
 
     /* Add a property to this propstat */
@@ -403,11 +445,11 @@ startelm(void *userdata, const struct ne_xml_elm *elm,
     /* Fill in the new property. */
     prop = &pstat->props[n];
 
-    prop->pname.name = prop->name = ne_strdup(elm->name);
-    if (elm->nspace[0] == '\0') {
+    prop->pname.name = prop->name = ne_strdup(name);
+    if (nspace[0] == '\0') {
 	prop->pname.nspace = prop->nspace = NULL;
     } else {
-	prop->pname.nspace = prop->nspace = ne_strdup(elm->nspace);
+	prop->pname.nspace = prop->nspace = ne_strdup(nspace);
     }
     prop->value = NULL;
 
@@ -427,27 +469,29 @@ startelm(void *userdata, const struct ne_xml_elm *elm,
 	prop->lang = NULL;
     }
 
-    return 0;
+    hdl->depth = 0;
+
+    return ELM_flatprop;
 }
 
-static int 
-endelm(void *userdata, const struct ne_xml_elm *elm, const char *cdata)
+static int endelm(void *userdata, int state,
+                  const char *nspace, const char *name)
 {
     ne_propfind_handler *hdl = userdata;
     struct propstat *pstat = ne_207_get_current_propstat(hdl->parser207);
     int n;
 
-    if (pstat == NULL) {
-	NE_DEBUG(NE_DBG_XML, "gp_endelm: No propstat found, or not my element.");
-	return -1;
+    if (hdl->depth > 0) {
+        /* nested. */
+        if (hdl->value->used < MAX_FLATPROP_LEN)
+            ne_buffer_concat(hdl->value, "</", nspace, name, ">", NULL);
+        hdl->depth--;
+    } else {
+        /* end of the current property value */
+        n = pstat->numprops - 1;
+        pstat->props[n].value = ne_buffer_finish(hdl->value);
+        hdl->value = ne_buffer_create();
     }
-
-    n = pstat->numprops - 1;
-
-    NE_DEBUG(NE_DBG_XML, "Value of property #%d is %s\n", n, cdata);
-    
-    pstat->props[n].value = ne_strdup(cdata);
-
     return 0;
 }
 
@@ -476,7 +520,7 @@ static void end_propstat(void *userdata, void *pstat_v,
 	int n;
 	
 	for (n = 0; n < pstat->numprops; n++) {
-	    free(pstat->props[n].value);
+	    ne_free(pstat->props[n].value);
 	    pstat->props[n].value = NULL;
 	}
     }
@@ -487,30 +531,38 @@ static void end_propstat(void *userdata, void *pstat_v,
 }
 
 /* Frees up a results set */
-static void free_propset(ne_prop_result_set *set)
+static void free_propset(ne_propfind_handler *handler,
+                         ne_prop_result_set *set)
 {
     int n;
     
+    if (handler->destructor && set->private) {
+        handler->destructor(handler->cd_userdata, set->private);
+    }
+
     for (n = 0; n < set->numpstats; n++) {
 	int m;
 	struct propstat *p = &set->pstats[n];
 
 	for (m = 0; m < p->numprops; m++) {
-	    NE_FREE(p->props[m].nspace);
-	    free(p->props[m].name);
-	    NE_FREE(p->props[m].lang);
-	    NE_FREE(p->props[m].value);
+            if (p->props[m].nspace) ne_free(p->props[m].nspace);
+            ne_free(p->props[m].name);
+            if (p->props[m].lang) ne_free(p->props[m].lang);
+            if (p->props[m].value) ne_free(p->props[m].value);
+            p->props[m].nspace = p->props[m].lang = 
+                p->props[m].value = NULL;
 	}
 
 	if (p->status.reason_phrase)
-	    free(p->status.reason_phrase);
+	    ne_free(p->status.reason_phrase);
 	if (p->props)
-	    free(p->props);
+	    ne_free(p->props);
     }
 
     if (set->pstats)
-	free(set->pstats);
-    free(set);
+	ne_free(set->pstats);
+    ne_uri_free(&set->uri);
+    ne_free(set);
 }
 
 static void end_response(void *userdata, void *resource,
@@ -519,31 +571,31 @@ static void end_response(void *userdata, void *resource,
 {
     ne_propfind_handler *handler = userdata;
     ne_prop_result_set *set = resource;
-    
-    /* TODO: Handle status here too? The status element is mandatory
-     * inside each propstat, so, not much point probably. */
 
     /* Pass back the results for this resource. */
-    if (handler->callback != NULL) {
-	handler->callback(handler->userdata, set->href, set);
-    }
-
-    free(set->href);
+    if (handler->callback && set->numpstats > 0)
+	handler->callback(handler->userdata, &set->uri, set);
 
     /* Clean up the propset tree we've just built. */
-    free_propset(set);
+    free_propset(handler, set);
+    handler->current = NULL;
 }
 
 ne_propfind_handler *
 ne_propfind_create(ne_session *sess, const char *uri, int depth)
 {
     ne_propfind_handler *ret = ne_calloc(sizeof(ne_propfind_handler));
+    ne_uri base = {0};
+
+    ne_fill_server_uri(sess, &base);
+    base.path = ne_strdup(uri);
 
     ret->parser = ne_xml_create();
-    ret->parser207 = ne_207_create(ret->parser, ret);
+    ret->parser207 = ne_207_create(ret->parser, &base, ret);
     ret->sess = sess;
     ret->body = ne_buffer_create();
     ret->request = ne_request_create(sess, "PROPFIND", uri);
+    ret->value = ne_buffer_create();
 
     ne_add_depth_header(ret->request, depth);
 
@@ -553,10 +605,15 @@ ne_propfind_create(ne_session *sess, const char *uri, int depth)
     ne_207_set_propstat_handlers(ret->parser207, start_propstat,
 				  end_propstat);
 
+    if (ne_get_session_flag(sess, NE_SESSFLAG_SHAREPOINT))
+        ne_207_set_flags(ret->parser207, NE_207_MSSP_ESCAPING);
+
     /* The start of the request body is fixed: */
-    ne_buffer_concat(ret->body, 
-		    "<?xml version=\"1.0\" encoding=\"utf-8\"?>" EOL 
-		    "<propfind xmlns=\"DAV:\">", NULL);
+    ne_buffer_czappend(ret->body, 
+                       "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" 
+                       "<propfind xmlns=\"DAV:\">");
+
+    ne_uri_free(&base);
 
     return ret;
 }
@@ -564,11 +621,14 @@ ne_propfind_create(ne_session *sess, const char *uri, int depth)
 /* Destroy a propfind handler */
 void ne_propfind_destroy(ne_propfind_handler *handler)
 {
+    ne_buffer_destroy(handler->value);
+    if (handler->current)
+        free_propset(handler, handler->current);
     ne_207_destroy(handler->parser207);
     ne_xml_destroy(handler->parser);
     ne_buffer_destroy(handler->body);
     ne_request_destroy(handler->request);
-    free(handler);    
+    ne_free(handler);    
 }
 
 int ne_simple_propfind(ne_session *sess, const char *href, int depth,
@@ -598,7 +658,7 @@ int ne_propnames(ne_session *sess, const char *href, int depth,
 
     hdl = ne_propfind_create(sess, href, depth);
 
-    ne_buffer_zappend(hdl->body, "<propname/></propfind>");
+    ne_buffer_czappend(hdl->body, "<propname/></propfind>");
 
     ret = propfind(hdl, results, userdata);
 
@@ -608,9 +668,11 @@ int ne_propnames(ne_session *sess, const char *href, int depth,
 }
 
 void ne_propfind_set_private(ne_propfind_handler *hdl,
-			      ne_props_create_complex creator,
-			      void *userdata)
+                             ne_props_create_complex creator,
+                             ne_props_destroy_complex destructor,
+                             void *userdata)
 {
-    hdl->private_creator = creator;
-    hdl->private_userdata = userdata;
+    hdl->creator = creator;
+    hdl->destructor = destructor;
+    hdl->cd_userdata = userdata;
 }
