@@ -1,6 +1,6 @@
 /* 
    HTTP request/response handling
-   Copyright (C) 1999-2009, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2010, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -190,7 +190,7 @@ static int aborted(ne_request *req, const char *doing, ssize_t code)
 
     switch(code) {
     case NE_SOCK_CLOSED:
-	if (sess->use_proxy) {
+	if (sess->nexthop->proxy != PROXY_NONE) {
 	    ne_set_error(sess, _("%s: connection was closed by proxy server"),
 			 doing);
 	} else {
@@ -284,6 +284,8 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
     ne_request *req = userdata;
 
     if (count) {
+        ssize_t ret;
+
         if (req->body.file.remain == 0)
             return 0;
 
@@ -292,7 +294,26 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
          * and 64-bit off64_t: */
         if ((ne_off_t)count > req->body.file.remain)
             count = (size_t)req->body.file.remain;
-	return read(req->body.file.fd, buffer, count);
+        
+        ret = read(req->body.file.fd, buffer, count);
+        if (ret > 0) {
+            req->body.file.remain -= ret;
+            return ret;
+        }
+        else if (ret == 0) {
+            ne_set_error(req->session, 
+                         _("Premature EOF in request body file"));
+        }
+        else if (ret < 0) {
+            char err[200];
+            int errnum = errno;
+
+            ne_set_error(req->session, 
+                         _("Failed reading request body file: %s"),
+                         ne_strerror(errnum, err, sizeof err));
+        }
+
+        return -1;
     } else {
         ne_off_t newoff;
 
@@ -308,7 +329,7 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
                 /* errno was set */
                 ne_strerror(errno, err, sizeof err);
             } else {
-                strcpy(err, _("offset invalid"));
+                ne_strnzcpy(err, _("offset invalid"), sizeof err);
             }
             ne_snprintf(offstr, sizeof offstr, "%" FMT_NE_OFF_T,
                         req->body.file.offset);
@@ -336,6 +357,14 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
 ((((code) == NE_SOCK_CLOSED || (code) == NE_SOCK_RESET || \
  (code) == NE_SOCK_TRUNC) && retry) ? NE_RETRY : (acode))
 
+/* For sending chunks, an 8-byte prefix is reserved at the beginning
+ * of the buffer.  This is large enough for a trailing \r\n for the
+ * previous chunk, the chunk size, and the \r\n following the
+ * chunk-size. */
+#define CHUNK_OFFSET (8)
+#define CHUNK_TERM "\r\n0\r\n\r\n"
+#define CHUNK_NULL_TERM "0\r\n\r\n"
+
 /* Sends the request body; returns 0 on success or an NE_* error code.
  * If retry is non-zero; will return NE_RETRY on persistent connection
  * timeout.  On error, the session error string is set and the
@@ -343,13 +372,28 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
 static int send_request_body(ne_request *req, int retry)
 {
     ne_session *const sess = req->session;
-    char buffer[NE_BUFSIZ];
+    char buffer[NE_BUFSIZ], *start;
     ssize_t bytes;
+    size_t buflen;
+    int chunked = req->body_length < 0, chunknum = 0;
+    int ret;
 
     NE_DEBUG(NE_DBG_HTTP, "Sending request body:\n");
 
+    /* Set up status union and (start, buflen) as the buffer to be
+     * passed the supplied callback. */
+    if (chunked) {
+        start = buffer + CHUNK_OFFSET;
+        buflen = sizeof(buffer) - CHUNK_OFFSET;
+        req->session->status.sr.total = -1;
+    }
+    else {
+        start = buffer;
+        buflen = sizeof buffer;
+        req->session->status.sr.total = req->body_length;
+    }
+
     req->session->status.sr.progress = 0;
-    req->session->status.sr.total = req->body_length;
     notify_status(sess, ne_status_sending);
     
     /* tell the source to start again from the beginning. */
@@ -358,8 +402,23 @@ static int send_request_body(ne_request *req, int retry)
         return NE_ERROR;
     }
     
-    while ((bytes = req->body_cb(req->body_ud, buffer, sizeof buffer)) > 0) {
-	int ret = ne_sock_fullwrite(sess->socket, buffer, bytes);
+    while ((bytes = req->body_cb(req->body_ud, start, buflen)) > 0) {
+        req->session->status.sr.progress += bytes;
+        if (chunked) {
+            /* Overwrite the buffer prefix with the appropriate chunk
+             * size; since ne_snprintf always NUL-terminates, the \n
+             * is omitted and placed over the NUL afterwards. */
+            if (chunknum++ == 0)
+                ne_snprintf(buffer, CHUNK_OFFSET, 
+                            "%06x\r", (unsigned)bytes);
+            else
+                ne_snprintf(buffer, CHUNK_OFFSET, 
+                            "\r\n%04x\r", (unsigned)bytes);
+            buffer[CHUNK_OFFSET - 1] = '\n';
+            bytes += CHUNK_OFFSET;
+        }
+        ret = ne_sock_fullwrite(sess->socket, buffer, bytes);
+
         if (ret < 0) {
             int aret = aborted(req, _("Could not send request body"), ret);
             return RETRY_RET(retry, ret, aret);
@@ -370,52 +429,69 @@ static int send_request_body(ne_request *req, int retry)
 		 bytes, (int)bytes, buffer);
 
         /* invoke progress callback */
-        req->session->status.sr.progress += bytes;
         notify_status(sess, ne_status_sending);
     }
 
-    if (bytes == 0) {
-        return NE_OK;
-    } else {
+    if (bytes) {
         NE_DEBUG(NE_DBG_HTTP, "Request body provider failed with "
                  "%" NE_FMT_SSIZE_T "\n", bytes);
         ne_close_connection(sess);
         return NE_ERROR;
     }
+
+    if (chunked) {
+        if (chunknum == 0)
+            ret = ne_sock_fullwrite(sess->socket, CHUNK_NULL_TERM, 
+                                    sizeof(CHUNK_NULL_TERM) - 1);
+        else
+            ret = ne_sock_fullwrite(sess->socket, CHUNK_TERM, 
+                                    sizeof(CHUNK_TERM) - 1);
+        if (ret < 0) {
+            int aret = aborted(req, _("Could not send chunked "
+                                      "request terminator"), ret);
+            return RETRY_RET(retry, ret, aret);
+        }
+    }
+    
+    return NE_OK;
 }
 
-/* Lob the User-Agent, connection and host headers in to the request
- * headers */
-static void add_fixed_headers(ne_request *req) 
+/* Set up buffer for initial request headers. */
+static ne_buffer *initial_request_headers(ne_request *req) 
 {
-    if (req->session->user_agent) {
-        ne_buffer_zappend(req->headers, req->session->user_agent);
+    ne_session *const sess = req->session;
+    ne_buffer *hdrs = ne_buffer_create();
+
+    if (sess->user_agent) {
+        ne_buffer_zappend(hdrs, sess->user_agent);
     }
 
     /* If persistent connections are disabled, just send Connection:
      * close; otherwise, send Connection: Keep-Alive to pre-1.1 origin
      * servers to try harder to get a persistent connection, except if
      * using a proxy as per 2068§19.7.1.  Always add TE: trailers. */
-    if (!req->session->flags[NE_SESSFLAG_PERSIST]) {
-       ne_buffer_czappend(req->headers,
-                          "Connection: TE, close" EOL
-                          "TE: trailers" EOL);
-    } else if (!req->session->is_http11 && !req->session->use_proxy) {
-        ne_buffer_czappend(req->headers, 
-                          "Keep-Alive: " EOL
-                          "Connection: TE, Keep-Alive" EOL
-                          "TE: trailers" EOL);
-    } else if (!req->session->is_http11 && req->session->use_proxy) {
-        ne_buffer_czappend(req->headers, 
+    if (!sess->flags[NE_SESSFLAG_PERSIST]) {
+       ne_buffer_czappend(hdrs, "Connection: TE, close" EOL);
+    } 
+    else if (!sess->is_http11 && !sess->any_proxy_http) {
+        ne_buffer_czappend(hdrs, 
+                           "Keep-Alive: " EOL
+                          "Connection: TE, Keep-Alive" EOL);
+    } 
+    else if (!req->session->is_http11 && !sess->any_proxy_http) {
+        ne_buffer_czappend(hdrs, 
                            "Keep-Alive: " EOL
                            "Proxy-Connection: Keep-Alive" EOL
-                           "Connection: TE" EOL
-                           "TE: trailers" EOL);
-    } else {
-        ne_buffer_czappend(req->headers, 
-                           "Connection: TE" EOL
-                           "TE: trailers" EOL);
+                           "Connection: TE" EOL);
+    } 
+    else {
+        ne_buffer_czappend(hdrs, "Connection: TE" EOL);
     }
+
+    ne_buffer_concat(hdrs, "TE: trailers" EOL "Host: ", 
+                     req->session->server.hostport, EOL, NULL);
+
+    return hdrs;
 }
 
 int ne_accept_always(void *userdata, ne_request *req, const ne_status *st)
@@ -434,23 +510,24 @@ ne_request *ne_request_create(ne_session *sess,
     ne_request *req = ne_calloc(sizeof *req);
 
     req->session = sess;
-    req->headers = ne_buffer_create();
     
     /* Presume the method is idempotent by default. */
     req->flags[NE_REQFLAG_IDEMPOTENT] = 1;
+    /* Expect-100 default follows the corresponding session flag. */
+    req->flags[NE_REQFLAG_EXPECT100] = sess->flags[NE_SESSFLAG_EXPECT100];
 
     /* Add in the fixed headers */
-    add_fixed_headers(req);
+    req->headers = initial_request_headers(req);
 
     /* Set the standard stuff */
     req->method = ne_strdup(method);
     req->method_is_head = (strcmp(method, "HEAD") == 0);
 
-    /* Only use an absoluteURI here when absolutely necessary: some
-     * servers can't parse them. */
-    if (req->session->use_proxy && !req->session->use_ssl && path[0] == '/')
+    /* Only use an absoluteURI here when we might be using an HTTP
+     * proxy, and SSL is in use: some servers can't parse them. */
+    if (sess->any_proxy_http && !req->session->use_ssl && path[0] == '/')
 	req->uri = ne_concat(req->session->scheme, "://", 
-			     req->session->server.hostport, path, NULL);
+                             req->session->server.hostport, path, NULL);
     else
 	req->uri = ne_strdup(path);
 
@@ -459,7 +536,7 @@ ne_request *ne_request_create(ne_session *sess,
 
 	for (hk = sess->create_req_hooks; hk != NULL; hk = hk->next) {
 	    ne_create_request_fn fn = (ne_create_request_fn)hk->fn;
-	    fn(req, hk->userdata, method, req->uri);
+	    fn(req, hk->userdata, req->method, req->uri);
 	}
     }
 
@@ -470,7 +547,12 @@ ne_request *ne_request_create(ne_session *sess,
 static void set_body_length(ne_request *req, ne_off_t length)
 {
     req->body_length = length;
-    ne_print_request_header(req, "Content-Length", "%" FMT_NE_OFF_T, length);
+
+    if (length >= 0)
+        ne_print_request_header(req, "Content-Length", "%" FMT_NE_OFF_T, length);
+    else /* length < 0 => chunked body */
+        ne_add_request_header(req, "Transfer-Encoding", "chunked");
+
 }
 
 void ne_set_request_body_buffer(ne_request *req, const char *buffer,
@@ -504,14 +586,14 @@ void ne_set_request_body_fd(ne_request *req, int fd,
 
 void ne_set_request_flag(ne_request *req, ne_request_flag flag, int value)
 {
-    if (flag < NE_SESSFLAG_LAST) {
+    if (flag < (ne_request_flag)NE_REQFLAG_LAST) {
         req->flags[flag] = value;
     }
 }
 
 int ne_get_request_flag(ne_request *req, ne_request_flag flag)
 {
-    if (flag < NE_REQFLAG_LAST) {
+    if (flag < (ne_request_flag)NE_REQFLAG_LAST) {
         return req->flags[flag];
     }
     return -1;
@@ -802,14 +884,13 @@ static ne_buffer *build_request(ne_request *req)
     struct hook *hk;
     ne_buffer *buf = ne_buffer_create();
 
-    /* Add Request-Line and Host header: */
-    ne_buffer_concat(buf, req->method, " ", req->uri, " HTTP/1.1" EOL,
-		     "Host: ", req->session->server.hostport, EOL, NULL);
-    
+    /* Add Request-Line and headers: */
+    ne_buffer_concat(buf, req->method, " ", req->uri, " HTTP/1.1" EOL, NULL);
+
     /* Add custom headers: */
     ne_buffer_append(buf, req->headers->data, ne_buffer_size(req->headers));
 
-    if (req->flags[NE_REQFLAG_EXPECT100]) {
+    if (req->body_length && req->flags[NE_REQFLAG_EXPECT100]) {
         ne_buffer_czappend(buf, "Expect: 100-continue\r\n");
     }
 
@@ -860,6 +941,12 @@ static inline void strip_eol(char *buf, ssize_t *len)
     }
 }
 
+#ifdef NE_HAVE_SSL
+#define SSL_CC_REQUESTED(_r) (_r->session->ssl_cc_requested)
+#else
+#define SSL_CC_REQUESTED(_r) (0)
+#endif
+
 /* Read and parse response status-line into 'status'.  'retry' is non-zero
  * if an NE_RETRY should be returned if an EOF is received. */
 static int read_status_line(ne_request *req, ne_status *status, int retry)
@@ -869,8 +956,11 @@ static int read_status_line(ne_request *req, ne_status *status, int retry)
 
     ret = ne_sock_readline(req->session->socket, buffer, sizeof req->respbuf);
     if (ret <= 0) {
-	int aret = aborted(req, _("Could not read status line"), ret);
-	return RETRY_RET(retry, ret, aret);
+        const char *errstr = SSL_CC_REQUESTED(req)
+            ? _("Could not read status line (TLS client certificate was requested)")
+            : _("Could not read status line");
+        int aret = aborted(req, errstr, ret);
+        return RETRY_RET(retry, ret, aret);
     }
     
     NE_DEBUG(NE_DBG_HTTP, "[status-line] < %s", buffer);
@@ -941,7 +1031,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
 	return RETRY_RET(retry, sret, aret);
     }
     
-    if (!req->flags[NE_REQFLAG_EXPECT100] && req->body_length > 0) {
+    if (!req->flags[NE_REQFLAG_EXPECT100] && req->body_length) {
 	/* Send request body, if not using 100-continue. */
 	ret = send_request_body(req, retry);
 	if (ret) {
@@ -961,7 +1051,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
 	if ((ret = discard_headers(req)) != NE_OK) break;
 
 	if (req->flags[NE_REQFLAG_EXPECT100] && (status->code == 100)
-            && req->body_length > 0 && !sentbody) {
+            && req->body_length && !sentbody) {
 	    /* Send the body after receiving the first 100 Continue */
 	    if ((ret = send_request_body(req, 0)) != NE_OK) break;	    
 	    sentbody = 1;
@@ -1119,11 +1209,9 @@ static int read_response_headers(ne_request *req)
 }
 
 /* Perform any necessary DNS lookup for the host given by *info;
- * return NE_ code. */
+ * returns NE_ code with error string set on error. */
 static int lookup_host(ne_session *sess, struct host_info *info)
 {
-    if (sess->addrlist) return NE_OK;
-
     NE_DEBUG(NE_DBG_HTTP, "Doing DNS lookup on %s...\n", info->hostname);
     sess->status.lu.hostname = info->hostname;
     notify_status(sess, ne_status_lookup);
@@ -1223,7 +1311,7 @@ int ne_begin_request(ne_request *req)
      * a) it is *necessary* to do so due to the use of a connection-auth
      * scheme, and
      * b) connection closure was not forced via "Connection: close".  */
-    if (req->session->use_proxy && !req->session->is_http11
+    if (req->session->nexthop->proxy == PROXY_HTTP && !req->session->is_http11
         && !forced_closure && req->session->flags[NE_SESSFLAG_CONNAUTH]) {
         value = get_response_header_hv(req, HH_HV_PROXY_CONNECTION,
                                        "proxy-connection");
@@ -1432,40 +1520,32 @@ static int proxy_tunnel(ne_session *sess)
 #endif
 
 /* Return the first resolved address for the given host. */
-static const ne_inet_addr *resolve_first(ne_session *sess, 
-                                         struct host_info *host)
+static const ne_inet_addr *resolve_first(struct host_info *host)
 {
-    if (sess->addrlist) {
-        sess->curaddr = 0;
-        return sess->addrlist[0];
-    } else {
-        return ne_addr_first(host->address);
-    }
+    return host->network ? host->network : ne_addr_first(host->address);
 }
 
 /* Return the next resolved address for the given host or NULL if
  * there are no more addresses. */
-static const ne_inet_addr *resolve_next(ne_session *sess,
-                                        struct host_info *host)
+static const ne_inet_addr *resolve_next(struct host_info *host)
 {
-    if (sess->addrlist) {
-        if (sess->curaddr++ < sess->numaddrs)
-            return sess->addrlist[sess->curaddr];
-        else
-            return NULL;
-    } else {
-        return ne_addr_next(host->address);
-    }
+    return host->network ? NULL : ne_addr_next(host->address);
 }
 
 /* Make new TCP connection to server at 'host' of type 'name'.  Note
  * that once a connection to a particular network address has
  * succeeded, that address will be used first for the next attempt to
  * connect. */
-static int do_connect(ne_session *sess, struct host_info *host, const char *err)
+static int do_connect(ne_session *sess, struct host_info *host)
 {
     int ret;
 
+    /* Resolve hostname if necessary. */
+    if (host->address == NULL && host->network == NULL) {
+        ret = lookup_host(sess, host);
+        if (ret) return ret;
+    }
+
     if ((sess->socket = ne_sock_create()) == NULL) {
         ne_set_error(sess, _("Could not create socket"));
         return NE_ERROR;
@@ -1478,7 +1558,7 @@ static int do_connect(ne_session *sess, struct host_info *host, const char *err)
         ne_sock_prebind(sess->socket, sess->local_addr, 0);
 
     if (host->current == NULL)
-	host->current = resolve_first(sess, host);
+	host->current = resolve_first(host);
 
     sess->status.ci.hostname = host->hostname;
 
@@ -1488,56 +1568,148 @@ static int do_connect(ne_session *sess, struct host_info *host, const char *err)
 #ifdef NE_DEBUGGING
 	if (ne_debug_mask & NE_DBG_HTTP) {
 	    char buf[150];
-	    NE_DEBUG(NE_DBG_HTTP, "Connecting to %s\n",
-		     ne_iaddr_print(host->current, buf, sizeof buf));
+	    NE_DEBUG(NE_DBG_HTTP, "req: Connecting to %s:%u\n",
+		     ne_iaddr_print(host->current, buf, sizeof buf),
+                     host->port);
 	}
 #endif
 	ret = ne_sock_connect(sess->socket, host->current, host->port);
     } while (ret && /* try the next address... */
-	     (host->current = resolve_next(sess, host)) != NULL);
+	     (host->current = resolve_next(host)) != NULL);
 
     if (ret) {
-        ne_set_error(sess, "%s: %s", err, ne_sock_error(sess->socket));
+        const char *msg;
+
+        if (host->proxy == PROXY_NONE)
+            msg = _("Could not connect to server");
+        else
+            msg = _("Could not connect to proxy server");
+
+        ne_set_error(sess, "%s: %s", msg, ne_sock_error(sess->socket));
         ne_sock_close(sess->socket);
 	return ret == NE_SOCK_TIMEOUT ? NE_TIMEOUT : NE_CONNECT;
     }
 
-    notify_status(sess, ne_status_connected);
-    
     if (sess->rdtimeout)
 	ne_sock_read_timeout(sess->socket, sess->rdtimeout);
 
+    notify_status(sess, ne_status_connected);
+    sess->nexthop = host;
+
     sess->connected = 1;
     /* clear persistent connection flag. */
     sess->persisted = 0;
     return NE_OK;
 }
 
+/* For a SOCKSv4 proxy only, the IP address of the origin server (in
+ * addition to the proxy) must be known, and must be an IPv4 address.
+ * Returns NE_*; connection closed and error string set on error. */
+static int socks_origin_lookup(ne_session *sess)
+{
+    const ne_inet_addr *ia;
+    int ret;
+
+    ret = lookup_host(sess, &sess->server);
+    if (ret) {
+        /* lookup_host already set the error string. */
+        ne_close_connection(sess);
+        return ret;
+    }
+    
+    /* Find the first IPv4 address available for the server. */
+    for (ia = ne_addr_first(sess->server.address);
+         ia && ne_iaddr_typeof(ia) == ne_iaddr_ipv6;
+         ia = ne_addr_next(sess->server.address)) {
+        /* noop */
+    }
+
+    /* ... if any */
+    if (ia == NULL) {
+        ne_set_error(sess, _("Could not find IPv4 address of "
+                             "hostname %s for SOCKS v4 proxy"), 
+                     sess->server.hostname);
+        ne_close_connection(sess);
+        return NE_LOOKUP;
+    }
+
+    sess->server.current = ia;
+    
+    return ret;
+}
+
 static int open_connection(ne_session *sess) 
 {
     int ret;
-    struct host_info *host;
     
     if (sess->connected) return NE_OK;
 
-    /* Resolve hostname if necessary. */
-    host = sess->use_proxy ? &sess->proxy : &sess->server;
-    if (host->address == NULL) {
-        ret = lookup_host(sess, host);
-        if (ret) return ret;
-    }    
-    
-    ret = do_connect(sess, host, 
-                     sess->use_proxy ? 
-                     _("Could not connect to proxy server")
-                     : _("Could not connect to server"));
-    if (ret != NE_OK) return ret;
+    if (!sess->proxies) {
+        ret = do_connect(sess, &sess->server);
+        if (ret) {
+            sess->nexthop = NULL;
+            return ret;
+        }
+    }
+    else {
+        struct host_info *hi;
+
+        /* Attempt to re-use proxy to avoid iterating through
+         * unnecessarily. */
+        if (sess->prev_proxy) 
+            ret = do_connect(sess, sess->prev_proxy);
+        else
+            ret = NE_ERROR;
+
+        /* Otherwise, try everything - but omitting prev_proxy if that
+         * has already been tried. */
+        for (hi = sess->proxies; hi && ret; hi = hi->next) {
+            if (hi != sess->prev_proxy)
+                ret = do_connect(sess, hi);
+        }
+
+        if (ret == NE_OK && sess->nexthop->proxy == PROXY_SOCKS) {
+            /* Special-case for SOCKS v4 proxies, which require the
+             * client to resolve the origin server IP address. */
+            if (sess->socks_ver == NE_SOCK_SOCKSV4) {
+                ret = socks_origin_lookup(sess);
+            }
+            
+            if (ret == NE_OK) {
+                /* Perform the SOCKS handshake, instructing the proxy
+                 * to set up the connection to the origin server. */
+                ret = ne_sock_proxy(sess->socket, sess->socks_ver, 
+                                    sess->server.current,
+                                    sess->server.hostname, sess->server.port,
+                                    sess->socks_user, sess->socks_password);
+                if (ret) {
+                    ne_set_error(sess, 
+                                 _("Could not establish connection from "
+                                   "SOCKS proxy (%s:%u): %s"),
+                                 sess->nexthop->hostname,
+                                 sess->nexthop->port,
+                                 ne_sock_error(sess->socket));
+                    ne_close_connection(sess);
+                    ret = NE_ERROR;
+                }
+            }
+        }
+
+        if (ret != NE_OK) {
+            sess->nexthop = NULL;
+            sess->prev_proxy = NULL;
+            return ret;
+        }
+        
+        /* Success - make this proxy stick. */
+        sess->prev_proxy = hi;
+    }
 
 #ifdef NE_HAVE_SSL
     /* Negotiate SSL layer if required. */
     if (sess->use_ssl && !sess->in_connect) {
-        /* CONNECT tunnel */
-        if (sess->use_proxy)
+        /* Set up CONNECT tunnel if using an HTTP proxy. */
+        if (sess->nexthop->proxy == PROXY_HTTP)
             ret = proxy_tunnel(sess);
         
         if (ret == NE_OK) {
