@@ -61,6 +61,8 @@ int test_argc;
 const char *test_suite;
 int test_num;
 
+static int quiet, count;
+
 /* statistics for all tests so far */
 static int passes = 0, fails = 0, skipped = 0, warnings = 0;
 
@@ -147,10 +149,32 @@ void in_child(void)
     flag_child = 1;
 }
 
+static const char dots[] = "......................";
+
+static void print_prefix(int n)
+{
+    if (quiet) {
+        printf("\r%s%.*s %2u/%2u ", test_suite, 
+               (int) (strlen(dots) - strlen(test_suite)), dots,
+               n + 1, count);
+    }
+    else {
+        if (warned) {
+	    printf("    %s ", dots);
+        }
+        else {
+            printf("\r%2d. %s%.*s ", n, test_name, 
+               (int) (strlen(dots) - strlen(test_name)), dots);
+        }
+    }
+    fflush(stdout);
+}
+
+
 int main(int argc, char *argv[])
 {
     int n;
-    static const char dots[] = "......................";
+    char *tmp;
     
     /* get basename(argv[0]) */
     test_suite = strrchr(argv[0], '/');
@@ -160,6 +184,9 @@ int main(int argc, char *argv[])
 	test_suite++;
     }
 
+    if (strncmp(test_suite, "lt-", 3) == 0)
+        test_suite += 3;
+
 #ifdef HAVE_SETLOCALE
     setlocale(LC_MESSAGES, "");
 #endif
@@ -215,8 +242,16 @@ int main(int argc, char *argv[])
 	printf(" Socket library initalization failed.\n");
     }
 
-    printf("-> running `%s':\n", test_suite);
+    if ((tmp = getenv("TEST_QUIET")) != NULL && strcmp(tmp, "1") == 0) {
+        quiet = 1;
+    }
+
+    if (!quiet)
+        printf("-> running `%s':\n", test_suite);
     
+    for (count = 0; tests[count].fn; count++)
+        /* nullop */;
+
     for (n = 0; !aborted && tests[n].fn != NULL; n++) {
 	int result, is_xfail = 0;
 #ifdef NEON_MEMLEAK
@@ -225,8 +260,9 @@ int main(int argc, char *argv[])
 #endif
 
 	test_name = tests[n].name;
-	printf("%2d. %s%.*s ", n, test_name, 
-	       (int) (strlen(dots) - strlen(test_name)), dots);
+
+        print_prefix(n);
+
 	have_context = 0;
 	test_num = n;
 	warned = 0;
@@ -267,38 +303,51 @@ int main(int argc, char *argv[])
             }
         }
 
-	/* align the result column if we've had warnings. */
-	if (warned) {
-	    printf("    %s ", dots);
-	}
+        print_prefix(n);
 
 	switch (result) {
 	case OK:
+	    passes++;
             if (is_xfail) {
                 COL("32;07"); 
-                printf("xfail");
-            } else {
+                printf("XFAIL");
+            } else if (!quiet) {
                 COL("32"); 
                 printf("pass"); 
             }
             NOCOL;
-	    if (warned) {
+            if (quiet && is_xfail) {
+                printf(" - %s", test_name);
+                if (have_context) {
+                    printf(" (%s)", test_context);
+                }
+            }
+	    if (warned && !quiet) {
 		printf(" (with %d warning%s)", warned, (warned > 1)?"s":"");
 	    }
 #ifdef NEON_MEMLEAK
             if (is_xleaky) {
-                printf(" (with expected leak, %" NE_FMT_SIZE_T " bytes)",
-                       ne_alloc_used - allocated);
+                if (quiet) {
+                    printf("expected leak - %s: %" NE_FMT_SIZE_T " bytes",
+                           test_name, ne_alloc_used - allocated);
+                }
+                else {
+                    printf(" (expected leak, %" NE_FMT_SIZE_T " bytes)",
+                           ne_alloc_used - allocated);
+                }
             }
 #endif
-	    putchar('\n');
-	    passes++;
+	    if (!quiet || is_xfail) putchar('\n');
 	    break;
 	case FAILHARD:
 	    aborted = 1;
+	    COL("41;37;01"); printf("fatal error - "); NOCOL;
 	    /* fall-through */
 	case FAIL:
 	    COL("41;37;01"); printf("FAIL"); NOCOL;
+            if (quiet) {
+                printf(" - %s", test_name);
+            }
 	    if (have_context) {
 		printf(" (%s)", test_context);
 	    }
@@ -310,6 +359,9 @@ int main(int argc, char *argv[])
 	    /* fall-through */
 	case SKIP:
 	    COL("44;37;01"); printf("SKIPPED"); NOCOL;
+            if (quiet) {
+                printf(" - %s", test_name);
+            }
 	    if (have_context) {
 		printf(" (%s)", test_context);
 	    }
@@ -323,30 +375,59 @@ int main(int argc, char *argv[])
 	}
 
 	reap_server();
+            
+        if (quiet) {
+            print_prefix(n);
+        }
     }
 
     /* discount skipped tests */
     if (skipped) {
-	printf("-> %d %s.\n", skipped,
-	       skipped==1?"test was skipped":"tests were skipped");
+        if (!quiet) 
+            printf("-> %d %s.\n", skipped,
+                   skipped == 1 ? "test was skipped" : "tests were skipped");
 	n -= skipped;
-	if (passes + fails != n) {
-	    printf("-> ARGH! Number of test results does not match "
-		   "number of tests.\n"
-		   "-> ARGH! Test Results are INRELIABLE.\n");
-	}
     }
     /* print the summary. */
     if (skipped && n == 0) {
-	printf("<- all tests skipped for `%s'.\n", test_suite);
+        if (quiet)
+            puts("(all skipped)");
+        else
+            printf("<- all tests skipped for `%s'.\n", test_suite);
     } else {
-	printf("<- summary for `%s': "
-	       "of %d tests run: %d passed, %d failed. %.1f%%\n",
-	       test_suite, n, passes, fails, 100*(float)passes/n);
+        if (quiet) {
+            printf("\r%s%.*s %2u/%2u ", test_suite, 
+                   (int) (strlen(dots) - strlen(test_suite)), dots,
+                   passes, count);
+            if (fails == 0) {
+                COL("32"); 
+                printf("passed");
+                NOCOL;
+                putchar(' ');
+            }
+            else {
+                printf("passed, %d failed ", fails);
+            }                       
+            if (skipped)
+                printf("(%d skipped) ", skipped);
+        }
+        else /* !quiet */
+            printf("<- summary for `%s': "
+                   "of %d tests run: %d passed, %d failed. %.1f%%\n",
+                   test_suite, n, passes, fails, 100*(float)passes/n);
 	if (warnings) {
-	    printf("-> %d warning%s issued.\n", warnings, 
-		   warnings==1?" was":"s were");
-	}
+	    if (quiet) {
+                printf("(%d warning%s)\n", warnings, 
+                       warnings > 1 ? "s" : "");
+            }
+            else {
+                printf("-> %d warning%s issued.\n", warnings, 
+                       warnings==1?" was":"s were");
+            }
+        }
+        else if (quiet) {
+            putchar('\n');
+        }
     }
 
     if (fclose(debug)) {
