@@ -1,6 +1,6 @@
 /* 
    lock tests
-   Copyright (C) 2002-2006, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2010, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -45,7 +45,7 @@
 static char *activelock(enum ne_lock_scope scope,
 			int depth,
 			const char *owner,
-			long timeout,
+			unsigned long timeout,
 			const char *token_href)
 {
     static char buf[BUFSIZ];
@@ -56,7 +56,7 @@ static char *activelock(enum ne_lock_scope scope,
 		"<D:lockscope><D:%s/></D:lockscope>\n"
 		"<D:depth>%d</D:depth>\n"
 		"<D:owner>%s</D:owner>\n"
-		"<D:timeout>Second-%ld</D:timeout>\n"
+		"<D:timeout>Second-%lu</D:timeout>\n"
 		"<D:locktoken><D:href>%s</D:href></D:locktoken>\n"
 		"</D:activelock>",
 		scope==ne_lockscope_exclusive?"exclusive":"shared",
@@ -69,15 +69,17 @@ static char *activelock(enum ne_lock_scope scope,
 static char *lock_response(enum ne_lock_scope scope,
 			   int depth,
 			   const char *owner,
-			   long timeout,
+			   unsigned long timeout,
 			   const char *token_href)
 {
     static char buf[BUFSIZ];
-    sprintf(buf, 
-	    "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
-	    "<D:prop xmlns:D=\"DAV:\">"
-	    "<D:lockdiscovery>%s</D:lockdiscovery></D:prop>\n",
-	    activelock(scope, depth, owner, timeout, token_href));
+
+    ne_snprintf(buf, sizeof buf,
+                "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
+                "<D:prop xmlns:D=\"DAV:\">"
+                "<D:lockdiscovery>%s</D:lockdiscovery></D:prop>\n",
+                activelock(scope, depth, owner, timeout, token_href));
+
     return buf;
 }
 
@@ -217,6 +219,14 @@ static int store_several(void)
     return OK;
 }
 
+/* Use a fake session forced to use port 7777 to the origin, to
+ * simplify the tests. */
+static int fake_session(ne_session **sess, server_fn fn, void *userdata)
+{
+    return proxied_session_server(sess, "http", "localhost", 7777,
+                                  fn, userdata);
+}
+
 /* regression test for <= 0.18.2, where timeout field was not parsed correctly. */
 static int lock_timeout(void)
 {
@@ -226,10 +236,11 @@ static int lock_timeout(void)
     struct ne_lock *lock = ne_lock_create();
 
     resp = ne_concat("HTTP/1.1 200 OK\r\n" "Server: neon-test-server\r\n"
+		     "Content-type: application/xml" EOL
 		     "Lock-Token: <opaquelocktoken:foo>" EOL
 		     "Connection: close\r\n\r\n", rbody, NULL);
 
-    CALL(make_session(&sess, single_serve_string, resp));
+    CALL(fake_session(&sess, single_serve_string, resp));
     ne_free(resp);
 
     ne_fill_server_uri(sess, &lock->uri);
@@ -249,6 +260,39 @@ static int lock_timeout(void)
     return OK;
 }
 
+#define LONG_TIMEOUT (4294967295UL)
+
+/* Lock timeouts should be allowed up to 2^32-1, but ne_lock uses a
+ * signed long to store timeouts, so this would fail with 32-bit long. */
+static int lock_long_timeout(void)
+{
+    ne_session *sess;
+    char *resp, *rbody = lock_response(ne_lockscope_exclusive, 0, "me",
+                                       LONG_TIMEOUT, "opaquelocktoken:foo");
+    struct ne_lock *lock = ne_lock_create();
+
+    resp = ne_concat("HTTP/1.1 200 OK\r\n" "Server: neon-test-server\r\n"
+                     "Content-type: application/xml" EOL
+                     "Lock-Token: <opaquelocktoken:foo>" EOL
+                     "Connection: close\r\n\r\n", rbody, NULL);
+
+    CALL(fake_session(&sess, single_serve_string, resp));
+    ne_free(resp);
+
+    ne_fill_server_uri(sess, &lock->uri);
+    lock->uri.path = ne_strdup("/foo");
+    lock->timeout = 5;
+
+    ONREQ(ne_lock(sess, lock));
+
+    ne_session_destroy(sess);
+    ne_lock_destroy(lock);
+
+    CALL(await_server());
+
+    return OK;
+}
+
 static int verify_if;
 static const char *verify_if_expect;
 
@@ -307,25 +351,36 @@ static int do_request(ne_session *sess, const char *path, int depth,
     return OK;
 }
 
+/* If modparent is non-zero; the request is flagged to
+ * modify the parent resource too. */
+#define LOCK_MODPARENT (0x01)
+/* Enable SharePoint hacks. */
+#define LOCK_SHAREPOINT (0x02)
+
 /* Tests If: header submission, for a lock of depth 'lockdepth' at
  * 'lockpath', with a request to 'reqpath' which Depth header of
- * 'reqdepth'.  If modparent is non-zero; the request is flagged to
- * modify the parent resource too. */
+ * 'reqdepth'.  'flags' is bitwise-or of LOCK_* flags above. */
 static int submit_test(const char *lockpath, int lockdepth,
 		       const char *reqpath, int reqdepth,
-		       int modparent)
+		       unsigned int flags)
 {
     ne_lock_store *store = ne_lockstore_create();
     ne_session *sess;
     struct ne_lock *lk = ne_lock_create();
     char *expect_if;
     int ret;
-    
-    expect_if = ne_concat("<http://localhost:7777", lockpath, 
-			  "> (<somelocktoken>)", NULL);
-    CALL(make_session(&sess, serve_verify_if, expect_if));
+
+    if (flags & LOCK_SHAREPOINT)
+        expect_if = ne_strdup("(<somelocktoken>)");
+    else
+        expect_if = ne_concat("<http://localhost:7777", lockpath,
+                              "> (<somelocktoken>)", NULL);
+    CALL(fake_session(&sess, serve_verify_if, expect_if));
     ne_free(expect_if);
 
+    if (flags & LOCK_SHAREPOINT)
+        ne_set_session_flag(sess, NE_SESSFLAG_SHAREPOINT, 1);
+
     ne_fill_server_uri(sess, &lk->uri);
     lk->uri.path = ne_strdup(lockpath);
     lk->token = ne_strdup("somelocktoken");
@@ -335,7 +390,7 @@ static int submit_test(const char *lockpath, int lockdepth,
     ne_lockstore_register(store, sess);
     ne_lockstore_add(store, lk);
 
-    ret = do_request(sess, reqpath, reqdepth, modparent);
+    ret = do_request(sess, reqpath, reqdepth, flags & LOCK_MODPARENT);
     CALL(await_server());
 
     ne_lockstore_destroy(store);
@@ -361,7 +416,7 @@ static int if_infinite_over(void)
 
 static int if_child(void)
 {
-    return submit_test("/foo/", 0, "/foo/bar", 0, 1);
+    return submit_test("/foo/", 0, "/foo/bar", 0, LOCK_MODPARENT);
 }
 
 /* this is a special test, where the PARENT resource of "/foo/bar" is
@@ -371,7 +426,13 @@ static int if_child(void)
  * correctly. */
 static int if_covered_child(void)
 {
-    return submit_test("/", NE_DEPTH_INFINITE, "/foo/bar", -1, 1);
+    return submit_test("/", NE_DEPTH_INFINITE, "/foo/bar", -1,
+                       LOCK_MODPARENT);
+}
+static int if_sharepoint(void)
+{
+    return submit_test("/foo-sharepoint", 0, "/foo-sharepoint", 0,
+                       LOCK_SHAREPOINT);
 }
 
 static int serve_discovery(ne_socket *sock, void *userdata)
@@ -418,7 +479,7 @@ static void discover_result(void *userdata, const struct ne_lock *lk,
 
 static int discover(void)
 {
-    ne_session *sess = ne_session_create("http", "localhost", 7777);
+    ne_session *sess;
     char *response;
     int ret;
     struct result_args args;
@@ -427,16 +488,17 @@ static int discover(void)
 
     args.lock->owner = ne_strdup("someowner");
     args.lock->token = ne_strdup("sometoken");
-
+    args.lock->uri.host = ne_strdup("localhost");
+    args.lock->uri.port = 7777;
+    args.lock->uri.scheme = ne_strdup("http");
+    
     /* default */
     args.result = FAIL;
     t_context("results callback never invoked");
 
-    ne_fill_server_uri(sess, &args.lock->uri);
-    args.lock->uri.path = ne_strdup("/lockme");
-
     response = discover_response("/lockme", args.lock);
-    CALL(spawn_server(7777, serve_discovery, response));
+    CALL(fake_session(&sess, serve_discovery, response));
+    args.lock->uri.path = ne_strdup("/lockme");
 
     ret = ne_lock_discover(sess, "/lockme", discover_result, &args);
     CALL(await_server());
@@ -472,11 +534,12 @@ static int lock_shared(void)
     rbody = multi_lock_response(resplocks);
     
     resp = ne_concat("HTTP/1.1 200 OK\r\n" "Server: neon-test-server\r\n"
+		     "Content-type: application/xml" EOL
 		     "Lock-Token: <opaquelocktoken:beta>" EOL
 		     "Connection: close\r\n\r\n", rbody, NULL);
     ne_free(rbody);
 
-    CALL(make_session(&sess, single_serve_string, resp));
+    CALL(fake_session(&sess, single_serve_string, resp));
     ne_free(resp);
 
     lock = ne_lock_create();
@@ -509,7 +572,7 @@ static int fail_discover(void)
     ne_session *sess;
     int ret;
     
-    CALL(make_session(&sess, single_serve_string, 
+    CALL(fake_session(&sess, single_serve_string, 
                       "HTTP/1.0 207 OK\r\n" "Connection: close\r\n" "\r\n"
                       "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
                       "<D:multistatus xmlns:D='DAV:'>\n"
@@ -530,7 +593,6 @@ static int no_creds(void *ud, const char *realm, int attempt,
 {
     return -1;
 }
-                    
 
 static int fail_lockauth(void)
 {
@@ -545,7 +607,7 @@ static int fail_lockauth(void)
         "\r\n";
     args.count = 2;
 
-    CALL(make_session(&sess, many_serve_string, &args));
+    CALL(fake_session(&sess, many_serve_string, &args));
 
     ne_set_server_auth(sess, no_creds, NULL);
 
@@ -570,6 +632,40 @@ static int fail_lockauth(void)
     return OK;
 }
 
+/* Regression test for neon 0.25.0 regression in ne_lock() error
+ * handling. */
+static int fail_noheader(void)
+{
+    ne_session *sess;
+    char *resp, *rbody = lock_response(ne_lockscope_exclusive, 0, "me",
+                                       6500, "opaquelocktoken:foo");
+    struct ne_lock *lock = ne_lock_create();
+    int ret;
+
+    resp = ne_concat("HTTP/1.1 200 OK\r\n" "Server: neon-test-server\r\n"
+                     "Content-type: application/xml" EOL
+                     "Connection: close\r\n\r\n", rbody, NULL);
+
+    CALL(fake_session(&sess, single_serve_string, resp));
+    ne_free(resp);
+
+    ne_fill_server_uri(sess, &lock->uri);
+    lock->uri.path = ne_strdup("/foo");
+    lock->timeout = NE_TIMEOUT_INFINITE;
+
+    ret = ne_lock(sess, lock);
+    ONN("LOCK request did not fail", ret != NE_ERROR);
+    
+    ONV(strstr(ne_get_error(sess), 
+               "LOCK response missing Lock-Token header") == NULL,
+        ("unexpected error: %s", ne_get_error(sess)));
+
+    ne_session_destroy(sess);
+    ne_lock_destroy(lock);
+
+    return await_server();
+}
+
 ne_test tests[] = {
     T(lookup_localhost),
     T(store_single),
@@ -579,11 +675,14 @@ ne_test tests[] = {
     T(if_infinite_over),
     T(if_child),
     T(if_covered_child),
+    T(if_sharepoint),
     T(lock_timeout),
+    T(lock_long_timeout),
     T(lock_shared),
     T(discover),
     T(fail_discover),
     T(fail_lockauth),
+    T(fail_noheader),
     T(NULL)
 };
 
