@@ -1,6 +1,6 @@
 /* 
    Tests for property handling
-   Copyright (C) 2002-2004, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2009, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,8 +36,7 @@
 #include "utils.h"
 
 static const ne_propname p_alpha = {"DAV:", "alpha"},
-    p_beta = {"http://webdav.org/random/namespace", "beta"},
-    p_delta = {NULL, "delta"};
+    p_beta = {"http://webdav.org/random/namespace", "beta"};
 
 /* Tests little except that ne_proppatch() doesn't segfault. */
 static int patch_simple(void)
@@ -59,7 +58,7 @@ static int patch_simple(void)
 
 #define RESP207 "HTTP/1.0 207 Stuff\r\n" "Server: foo\r\n\r\n"
 
-static void dummy_results(void *ud, const char *href,
+static void dummy_results(void *ud, const ne_uri *uri,
 			  const ne_prop_result_set *rset)
 {
     NE_DEBUG(NE_DBG_HTTP, "dummy_results.\n");
@@ -142,11 +141,11 @@ static int pstat_count;
 
 /* tos_*: set of 207 callbacks which serialize the data back into a
  * text stream, which can be easily checked for correctness. */
-static void *tos_startresp(void *buf, const char *href)
+static void *tos_startresp(void *buf, const ne_uri *uri)
 {
-    ne_buffer_concat(buf, "start-resp[", href, "];", NULL);
+    ne_buffer_concat(buf, "start-resp[", uri->path, "];", NULL);
     pstat_count = 0;
-    return ne_strdup(href);
+    return ne_strdup(uri->path);
 }
 
 static void tos_status_descr(ne_buffer *buf, const ne_status *status,
@@ -242,12 +241,20 @@ static int tos_endprop(void *userdata, int state,
 static int run_207_response(char *resp, const char *expected)
 {
     ne_buffer *buf = ne_buffer_create();
-    ne_session *sess = ne_session_create("http", "localhost", 7777);
+    ne_session *sess;
     ne_xml_parser *p = ne_xml_create();
-    ne_207_parser *p207 = ne_207_create(p, buf);
-    ne_request *req = ne_request_create(sess, "PROPFIND", "/foo");
+    ne_207_parser *p207;
+    ne_request *req;
+    ne_uri base = {0};
     struct propctx ctx;
 
+    CALL(session_server(&sess, single_serve_string, resp));
+    req = ne_request_create(sess, "PROPFIND", "/foo");
+    ne_fill_server_uri(sess, &base);
+    base.path = ne_strdup("/foo");
+    p207 = ne_207_create(p, &base, buf);
+    ne_uri_free(&base);
+
     ne_add_response_body_reader(req, ne_accept_207, ne_xml_parse_v, p);
 
     ne_207_set_response_handlers(p207, tos_startresp, tos_endresp);
@@ -257,7 +264,6 @@ static int run_207_response(char *resp, const char *expected)
     ctx.p207 = p207;
     ne_xml_push_handler(p, tos_startprop, tos_cdata, tos_endprop, &ctx);
 
-    CALL(spawn_server(7777, single_serve_string, resp));
 
     ONREQ(ne_request_dispatch(req));
 
@@ -405,10 +411,12 @@ static int two_oh_seven(void)
 }
 
 /* Serialize propfind result callbacks into a string */
-static int simple_iterator(void *buf, const ne_propname *name,
+static int simple_iterator(void *vbuf, const ne_propname *name,
                            const char *value, const ne_status *st)
 {
     char code[20];
+    ne_buffer *buf = vbuf;
+
     ne_buffer_concat(buf, "prop:[{", name->nspace, ",", 
                      name->name, "}=", NULL);
     if (value)
@@ -423,24 +431,31 @@ static int simple_iterator(void *buf, const ne_propname *name,
     return 0;
 }
 
-static void simple_results(void *buf, const char *href,
+static void simple_results(void *buf, const ne_uri *uri,
                            const ne_prop_result_set *rset)
 {
-    ne_buffer_concat(buf, "results(", href, ",", NULL);
+    ne_buffer_concat(buf, "results(", uri->path, ",", NULL);
     ne_propset_iterate(rset, simple_iterator, buf);
-    ne_buffer_zappend(buf, ")//");
+    ne_buffer_czappend(buf, ")//");
 }
 
+/* Test function to compare two long strings and print a digestible
+ * failure message. */
 static int diffcmp(const char *expected, const char *actual)
 {
     size_t n;
     
     if (!strcmp(expected, actual)) return OK;
 
+    NE_DEBUG(NE_DBG_HTTP, 
+             "diffcmp: Expect: [%s]\n"
+             "diffcmp: Actual: [%s]\n",
+             expected, actual);
+
     for (n = 0; expected[n] && actual[n]; n++) {
         if (expected[n] != actual[n]) {
             t_context("difference at byte %" NE_FMT_SIZE_T ": "
-                      "`%.6s...' not `%.6s...'",
+                      "`%.10s...' not `%.10s...'",
                       n, actual+n, expected+n);
             break;
         }
@@ -448,25 +463,82 @@ static int diffcmp(const char *expected, const char *actual)
 
     return FAIL;
 }
-       
 
-static int run_simple_propfind(const ne_propname *props, char *resp, 
-                               int depth, const char *expected)
+/* PROPFIND creator callback. */
+static void *pf_creator(void *userdata, const ne_uri *uri)
 {
-    ne_session *sess = ne_session_create("http", "localhost", 7777);
+    ne_buffer *buf = userdata;
+
+    NE_DEBUG(NE_DBG_HTTP, "pf: Creator at %s\n", uri->path);
+
+    ne_buffer_concat(buf, "creator[", uri->path, "]//", NULL);
+
+    return ne_strdup(uri->path);
+}
+
+/* PROPFIND destructor callback. */
+static void pf_destructor(void *userdata, void *private)
+{
+    ne_buffer *buf = userdata;
+    char *cookie = private;
+
+    NE_DEBUG(NE_DBG_HTTP, "pf: Destructor at %s\n", cookie);
+
+    ne_buffer_concat(buf, "destructor[", cookie, "]//", NULL);
+    
+    ne_free(cookie);
+}
+
+  
+/* PROPFIND test type. */     
+enum pftype { 
+    PF_SIMPLE, /* using ne_simple_propfind */
+    PF_NAMED,  /* using ne_propfind_named */
+    PF_SP_NAMED, /* using ne_propfind_named w/SHAREPOINT hacks */
+    PF_ALLPROP /* using ne_propfind_allprop */
+};
+
+static int run_propfind(const ne_propname *props, char *resp, 
+                        int depth, const char *expected, enum pftype type)
+{
+    ne_session *sess;
     ne_buffer *buf = ne_buffer_create();
 
-    CALL(spawn_server(7777, single_serve_string, resp));
+    CALL(make_session(&sess, single_serve_string, resp));
 
-    ONREQ(ne_simple_propfind(sess, "/propfind", depth, props,
-                             simple_results, buf));
+    if (type == PF_SIMPLE) {
+        ONREQ(ne_simple_propfind(sess, "/propfind", depth, props,
+                                 simple_results, buf));
+    }
+    else {
+        ne_propfind_handler *hdl;
 
+        if (type == PF_SP_NAMED) {
+            ne_set_session_flag(sess, NE_SESSFLAG_SHAREPOINT, 1);
+            type = PF_NAMED;
+        }
+
+        hdl = ne_propfind_create(sess, "/propfind", depth);
+
+        ne_propfind_set_private(hdl, pf_creator, pf_destructor,
+                                buf);
+        
+        if (type == PF_NAMED) {
+            ONREQ(ne_propfind_named(hdl, props, simple_results, buf));
+        }
+        else {
+            ONREQ(ne_propfind_allprop(hdl, simple_results, buf));
+        }
+
+        ne_propfind_destroy(hdl);
+    }
+
+    ne_session_destroy(sess);
     CALL(await_server());
 
     CALL(diffcmp(expected, buf->data));
 
     ne_buffer_destroy(buf);
-    ne_session_destroy(sess);
     return OK;
 }
 
@@ -476,17 +548,18 @@ static int run_simple_propfind(const ne_propname *props, char *resp,
 	PSTAT_207(PROPS_207(APROP_207("fishbone", value)) \
 	STAT_207(status))))
 
-static int pfind_simple(void)
+static int propfind(void)
 {
     static const struct {
         char *resp;
         const char *expected;
-        int depth, pset;
+        int depth;
+        enum pftype type;
     } ts[] = {
         /* simple single property. */
         { FISHBONE_RESP("hello, world", "212 Well OK"),
           "results(/foop,prop:[{DAV:,fishbone}='hello, world':{212 Well OK}];)//",
-          0, 0 },
+          0, PF_SIMPLE },
         /* property with some nested elements. */
         { FISHBONE_RESP("this is <foo/> a property <bar><lemon>fish</lemon></bar> value", 
                         "299 Just About OK"),
@@ -494,14 +567,14 @@ static int pfind_simple(void)
           "'this is <foo></foo> a property "
           "<bar><lemon>fish</lemon></bar> value':"
           "{299 Just About OK}];)//",
-          0, 0 },
+          0, PF_SIMPLE },
 
         /* failed to fetch a property. */
         { FISHBONE_RESP("property value is ignored", 
                         "404 Il n'ya pas de property"),
           "results(/foop,prop:[{DAV:,fishbone}=#novalue#:"
           "{404 Il n'ya pas de property}];)//",
-          0, 0 },
+          0, PF_SIMPLE },
 
 #if 0
         /* propstat missing status should be ignored; if a response contains no
@@ -509,7 +582,8 @@ static int pfind_simple(void)
         { MULTI_207(RESP_207("/alpha", PSTAT_207(APROP_207("fishbone", "unseen")))
                     RESP_207("/beta", PSTAT_207(APROP_207("fishbone", "hello, world")
                                                 STAT_207("200 OK")))),
-          "results(/beta,prop:[{DAV:,fishbone}='hello, world':{200 OK}];)//", 0, 0},
+          "results(/beta,prop:[{DAV:,fishbone}='hello, world':{200 OK}];)//", 0, 
+          PF_SIMPLE },
 #endif
 
         /* props on several resources */
@@ -521,14 +595,49 @@ static int pfind_simple(void)
                                        STAT_207("256 Second is OK")))),
           "results(/alpha,prop:[{DAV:,fishbone}='strike one':{234 First is OK}];)//"
           "results(/beta,prop:[{DAV:,fishbone}='strike two':{256 Second is OK}];)//",
-          0, 0},
+          0, PF_SIMPLE},
 
         /* whitespace handling. */
-        { MULTI_207(RESP_207("\r\nhttp://localhost:7777/alpha ",
+        { MULTI_207(RESP_207("\r\nhttp://localhost/alpha ",
                              PSTAT_207(PROPS_207(APROP_207("alpha", "beta"))
                                        "<D:status>\r\nHTTP/1.1 200 OK </D:status>"))),
-          "results(http://localhost:7777/alpha,prop:[{DAV:,alpha}='beta':{200 OK}];)//",
-          0, 0}
+          "results(/alpha,prop:[{DAV:,alpha}='beta':{200 OK}];)//",
+          0, PF_SIMPLE},
+        
+        /* attribute handling. */
+        { MULTI_207(RESP_207("\r\nhttp://localhost/alpha ",
+                             PSTAT_207(PROPS_207("<D:alpha>"
+                                                 "<D:foo D:fee='bar' bar='fee'>beta</D:foo></D:alpha>")
+                                       "<D:status>\r\nHTTP/1.1 200 OK </D:status>"))),
+          "results(/alpha,prop:[{DAV:,alpha}='<DAV:foo DAV:fee='bar' bar='fee'>beta</DAV:foo>':{200 OK}];)//",
+          0, PF_SIMPLE},
+        
+
+        /* "complex" propfinds. */
+
+        { FISHBONE_RESP("hello, world", "212 Well OK"),
+          "creator[/foop]//"
+          "results(/foop,prop:[{DAV:,fishbone}='hello, world':{212 Well OK}];)//"
+          "destructor[/foop]//",
+          0, PF_NAMED },
+
+        /* 207 with badly encoded URI in href */
+        { MULTI_207(RESP_207("http://example.com/foo€bar", \
+                             PSTAT_207(PROPS_207(APROP_207("fishbone", "hello, world")) \
+                                       STAT_207("209 Good News")))),
+          "creator[/foo%e2%82%acbar]//"
+          "results(/foo%e2%82%acbar,prop:[{DAV:,fishbone}='hello, world':{209 Good News}];)//"
+          "destructor[/foo%e2%82%acbar]//",
+          0, PF_SP_NAMED },
+
+        { MULTI_207(RESP_207("/foo%20bar/€bar", \
+                             PSTAT_207(PROPS_207(APROP_207("fishbone", "hello, world")) \
+                                       STAT_207("209 Good News")))),
+          "creator[/foo%20bar/%e2%82%acbar]//"
+          "results(/foo%20bar/%e2%82%acbar,prop:[{DAV:,fishbone}='hello, world':{209 Good News}];)//"
+          "destructor[/foo%20bar/%e2%82%acbar]//",
+          0, PF_SP_NAMED }
+
     };
     const ne_propname pset1[] = {
         { "DAV:", "fishbone", },
@@ -539,8 +648,8 @@ static int pfind_simple(void)
     for (n = 0; n < sizeof(ts)/sizeof(ts[0]); n++) {
         const ne_propname *pset = pset1;
 
-        CALL(run_simple_propfind(pset, ts[n].resp, ts[n].depth,
-                                 ts[n].expected));
+        CALL(run_propfind(pset, ts[n].resp, ts[n].depth,
+                          ts[n].expected, ts[n].type));
     }
 
 
@@ -551,12 +660,9 @@ static int unbounded_response(const char *header, const char *repeats)
 {
     ne_session *sess;
     struct infinite i = { header, repeats};
-    int dbg;
 
     CALL(make_session(&sess, serve_infinite, &i));
 
-    dbg = ne_debug_mask;
-
     ONN("unbounded PROPFIND response did not fail",
         ne_simple_propfind(sess, "/", 0, NULL, 
                            dummy_results, NULL) != NE_ERROR);
@@ -587,7 +693,7 @@ static int unbounded_props(void)
 ne_test tests[] = {
     T(two_oh_seven),
     T(patch_simple),
-    T(pfind_simple),
+    T(propfind),
     T(regress),
     T(patch_regress),
     T(unbounded_props),
