@@ -222,7 +222,6 @@ static const unsigned char raw_127[4] = "\x7f\0\0\01", /* 127.0.0.1 */
     raw6_nuls[16] = /* :: */ "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
 #ifdef TEST_IPV6
 static const unsigned char 
-raw6_local[16] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1",
 raw6_cafe[16] = /* feed::cafe */ "\xfe\xed\0\0\0\0\0\0\0\0\0\0\0\0\xca\xfe",
 raw6_babe[16] = /* cafe:babe:: */ "\xca\xfe\xba\xbe\0\0\0\0\0\0\0\0\0\0\0\0";
 #endif
@@ -374,6 +373,7 @@ static int addr_reverse(void)
 {
     ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
     char buf[128], *syshost = NULL;
+    int match;
 
 #ifdef HAVE_GETHOSTNAME
     char host[128];
@@ -388,10 +388,20 @@ static int addr_reverse(void)
     ONN("reverse lookup for 127.0.0.1 failed",
         ne_iaddr_reverse(ia, buf, sizeof buf) != 0);
 
-    ONV(!(strcmp(buf, "localhost.localdomain") == 0
-          || strcmp(buf, "localhost") == 0
-          || (syshost && strcmp(buf, syshost) == 0)),
-        ("reverse lookup for 127.0.0.1 got %s", buf));
+    NE_DEBUG(NE_DBG_SOCKET, "Reverse lookup for 127.0.0.1 => %s\n", buf);
+
+    match = strcmp(buf, "localhost.localdomain") == 0
+        || strcmp(buf, "localhost") == 0;
+
+    if (!match && syshost)
+        /* If the returned name has the system hostname as a prefix, that's
+         * good enough. */
+        match = strncmp(buf, syshost, strlen(syshost)) == 0;
+
+    if (!match)
+        t_warning("reverse lookup for 127.0.0.1 got '%s'", buf);
+    
+    ONN("reverse lookup for 127.0.0.1 got empty string", strlen(buf) == 0);
 
     ne_iaddr_free(ia);
 
@@ -1036,6 +1046,9 @@ static int ssl_closure(void)
 
 static int serve_truncate(ne_socket *sock, void *userdata)
 {
+    if (ne_sock_read(sock, buffer, 1) != 1)
+        NE_DEBUG(NE_DBG_SOCKET, "serve_truncate failed to read a byte.\n");
+    NE_DEBUG(NE_DBG_SOCKET, "read a byte, exiting...\n");
     exit(0);
 }
 
@@ -1046,6 +1059,7 @@ static int ssl_truncate(void)
     ne_socket *sock; int ret;
     
     CALL(begin(&sock, serve_truncate, NULL));
+    CALL(full_write(sock, "a", 1));
     ret = ne_sock_read(sock, buffer, 1);
     ONV(ret != NE_SOCK_TRUNC,
 	("socket got error %d not truncation: `%s'", ret,
